# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Any, Dict

from rdflib import Graph, Literal

from swh.indexer.codemeta import CROSSWALK_TABLE
from swh.indexer.namespaces import SCHEMA

from .base import SingleFileIntrinsicMapping, XmlMapping
from .utils import add_url_if_valid, prettyprint_graph  # noqa


class MavenMapping(XmlMapping, SingleFileIntrinsicMapping):
    """
    dedicated class for Maven (pom.xml) mapping and translation
    """

    name = "maven"
    filename = b"pom.xml"
    mapping = CROSSWALK_TABLE["Java (Maven)"]
    string_fields = ["name", "version", "description", "email"]

    _default_repository = {"url": "https://repo.maven.apache.org/maven2/"}

    def _translate_dict(self, d: Dict[str, Any]) -> Dict[str, Any]:
        return super()._translate_dict(d.get("project") or {})

    def extra_translation(self, graph: Graph, root, d):
        self.parse_repositories(graph, root, d)

    def parse_repositories(self, graph: Graph, root, d):
        """https://maven.apache.org/pom.html#Repositories

        >>> import rdflib
        >>> import xmltodict
        >>> from pprint import pprint
        >>> d = xmltodict.parse('''
        ... <repositories>
        ...   <repository>
        ...     <id>codehausSnapshots</id>
        ...     <name>Codehaus Snapshots</name>
        ...     <url>http://snapshots.maven.codehaus.org/maven2</url>
        ...     <layout>default</layout>
        ...   </repository>
        ... </repositories>
        ... ''')
        >>> MavenMapping().parse_repositories(rdflib.Graph(), rdflib.BNode(), d)
        """
        repositories = d.get("repositories")
        if not repositories:
            self.parse_repository(graph, root, d, self._default_repository)
        elif isinstance(repositories, dict):
            repositories = repositories.get("repository") or []
            if not isinstance(repositories, list):
                repositories = [repositories]
            for repo in repositories:
                self.parse_repository(graph, root, d, repo)

    def parse_repository(self, graph: Graph, root, d, repo):
        if not isinstance(repo, dict):
            return
        if repo.get("layout", "default") != "default":
            return  # TODO ?
        url = repo.get("url")
        group_id = d.get("groupId")
        artifact_id = d.get("artifactId")
        if (
            isinstance(url, str)
            and isinstance(group_id, str)
            and isinstance(artifact_id, str)
        ):
            repo = os.path.join(url, *group_id.split("."), artifact_id)
            if "${" in repo:
                # Often use as templating in pom.xml files collected from VCSs
                return
            add_url_if_valid(graph, root, SCHEMA.codeRepository, repo)

    def normalize_groupId(self, id_):
        """https://maven.apache.org/pom.html#Maven_Coordinates

        >>> MavenMapping().normalize_groupId('org.example')
        rdflib.term.Literal('org.example')
        """
        if isinstance(id_, str):
            return Literal(id_)

    def translate_licenses(self, graph, root, licenses):
        """https://maven.apache.org/pom.html#Licenses

        >>> import xmltodict
        >>> import json
        >>> from rdflib import URIRef
        >>> d = xmltodict.parse('''
        ... <licenses>
        ...   <license>
        ...     <name>Apache License, Version 2.0</name>
        ...     <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
        ...   </license>
        ... </licenses>
        ... ''')
        >>> print(json.dumps(d, indent=4))
        {
            "licenses": {
                "license": {
                    "name": "Apache License, Version 2.0",
                    "url": "https://www.apache.org/licenses/LICENSE-2.0.txt"
                }
            }
        }
        >>> graph = Graph()
        >>> root = URIRef("http://example.org/test-software")
        >>> MavenMapping().translate_licenses(graph, root, d["licenses"])
        >>> prettyprint_graph(graph, root)
        {
            "@id": ...,
            "http://schema.org/license": {
                "@id": "https://www.apache.org/licenses/LICENSE-2.0.txt"
            }
        }

        or, if there are more than one license:

        >>> import xmltodict
        >>> from pprint import pprint
        >>> d = xmltodict.parse('''
        ... <licenses>
        ...   <license>
        ...     <name>Apache License, Version 2.0</name>
        ...     <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
        ...   </license>
        ...   <license>
        ...     <name>MIT License</name>
        ...     <url>https://opensource.org/licenses/MIT</url>
        ...   </license>
        ... </licenses>
        ... ''')
        >>> graph = Graph()
        >>> root = URIRef("http://example.org/test-software")
        >>> MavenMapping().translate_licenses(graph, root, d["licenses"])
        >>> pprint(set(graph.triples((root, URIRef("http://schema.org/license"), None))))
        {(rdflib.term.URIRef('http://example.org/test-software'),
          rdflib.term.URIRef('http://schema.org/license'),
          rdflib.term.URIRef('https://opensource.org/licenses/MIT')),
         (rdflib.term.URIRef('http://example.org/test-software'),
          rdflib.term.URIRef('http://schema.org/license'),
          rdflib.term.URIRef('https://www.apache.org/licenses/LICENSE-2.0.txt'))}
        """

        if not isinstance(licenses, dict):
            return
        licenses = licenses.get("license")
        if isinstance(licenses, dict):
            licenses = [licenses]
        elif not isinstance(licenses, list):
            return
        for license in licenses:
            if isinstance(license, dict):
                add_url_if_valid(graph, root, SCHEMA.license, license.get("url"))



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import inspect

import pytest

from swh.indexer.storage import IndexerStorage, get_indexer_storage
from swh.indexer.storage.api.client import RemoteStorage
from swh.indexer.storage.in_memory import IndexerStorage as MemoryIndexerStorage
from swh.indexer.storage.interface import IndexerStorageInterface

SERVER_IMPLEMENTATIONS_KWARGS = [
    ("remote", RemoteStorage, {"url": "localhost"}),
    ("local", IndexerStorage, {"db": "something"}),
]

SERVER_IMPLEMENTATIONS = SERVER_IMPLEMENTATIONS_KWARGS + [
    ("memory", MemoryIndexerStorage, {}),
]


@pytest.fixture
def mock_psycopg2(mocker):
    mocker.patch("swh.indexer.storage.psycopg2.pool")
    return mocker


def test_init_get_indexer_storage_failure():
    with pytest.raises(ValueError, match="Unknown indexer storage class"):
        get_indexer_storage("unknown-idx-storage")


@pytest.mark.parametrize("class_name,expected_class,kwargs", SERVER_IMPLEMENTATIONS)
def test_init_get_indexer_storage(class_name, expected_class, kwargs, mock_psycopg2):
    if kwargs:
        concrete_idx_storage = get_indexer_storage(class_name, **kwargs)
    else:
        concrete_idx_storage = get_indexer_storage(class_name)
    assert isinstance(concrete_idx_storage, expected_class)
    assert isinstance(concrete_idx_storage, IndexerStorageInterface)


@pytest.mark.parametrize(
    "class_name,expected_class,kwargs", SERVER_IMPLEMENTATIONS_KWARGS
)
def test_init_get_indexer_storage_deprecation_warning(
    class_name, expected_class, kwargs, mock_psycopg2
):
    with pytest.warns(DeprecationWarning):
        concrete_idx_storage = get_indexer_storage(class_name, args=kwargs)
    assert isinstance(concrete_idx_storage, expected_class)


def test_types(swh_indexer_storage) -> None:
    """Checks all methods of StorageInterface are implemented by this
    backend, and that they have the same signature."""
    # Create an instance of the protocol (which cannot be instantiated
    # directly, so this creates a subclass, then instantiates it)
    interface = type("_", (IndexerStorageInterface,), {})()

    assert "content_mimetype_add" in dir(interface)

    missing_methods = []

    for meth_name in dir(interface):
        if meth_name.startswith("_"):
            continue
        interface_meth = getattr(interface, meth_name)
        try:
            concrete_meth = getattr(swh_indexer_storage, meth_name)
        except AttributeError:
            missing_methods.append(meth_name)
            continue

        expected_signature = inspect.signature(interface_meth)
        actual_signature = inspect.signature(concrete_meth)

        assert expected_signature == actual_signature, meth_name

    assert missing_methods == []

    # If all the assertions above succeed, then this one should too.
    # But there's no harm in double-checking.
    # And we could replace the assertions above by this one, but unlike
    # the assertions above, it doesn't explain what is missing.
    assert isinstance(swh_indexer_storage, IndexerStorageInterface)



# Copyright (C) 2017-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import logging
from typing import Any, Callable, Dict, List, Optional, Pattern, Tuple, TypeVar, Union
import uuid
import xml.parsers.expat

from pyld import jsonld
import rdflib
from typing_extensions import TypedDict
import xmltodict
import yaml

from swh.indexer.codemeta import _document_loader, compact
from swh.indexer.namespaces import RDF, SCHEMA
from swh.indexer.storage.interface import Sha1

from .utils import add_url_if_valid

TMP_ROOT_URI_PREFIX = "https://www.softwareheritage.org/schema/2022/indexer/tmp-node/"
"""Prefix used to generate temporary URIs for root nodes being translated."""


class DirectoryLsEntry(TypedDict):
    target: Sha1
    sha1: Optional[Sha1]
    name: bytes
    type: str


TTranslateCallable = TypeVar(
    "TTranslateCallable",
    bound=Callable[[Any, rdflib.Graph, rdflib.term.BNode, Any], None],
)


def produce_terms(*uris: str) -> Callable[[TTranslateCallable], TTranslateCallable]:
    """Returns a decorator that marks the decorated function as adding
    the given terms to the ``translated_metadata`` dict"""

    def decorator(f: TTranslateCallable) -> TTranslateCallable:
        if not hasattr(f, "produced_terms"):
            f.produced_terms = []  # type: ignore
        f.produced_terms.extend(uris)  # type: ignore
        return f

    return decorator


class BaseMapping:
    """Base class for :class:`BaseExtrinsicMapping` and :class:`BaseIntrinsicMapping`,
    not to be inherited directly."""

    def __init__(self, log_suffix=""):
        self.log_suffix = log_suffix
        self.log = logging.getLogger(
            "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        )

    @property
    def name(self):
        """A name of this mapping, used as an identifier in the
        indexer storage."""
        raise NotImplementedError(f"{self.__class__.__name__}.name")

    def translate(self, raw_content: bytes) -> Optional[Dict]:
        """
        Translates content by parsing content from a bytestring containing
        mapping-specific data and translating with the appropriate mapping
        to JSON-LD using the Codemeta and ForgeFed vocabularies.

        Args:
            raw_content: raw content to translate

        Returns:
            translated metadata in JSON friendly form needed for the content
            if parseable, :const:`None` otherwise.

        """
        raise NotImplementedError(f"{self.__class__.__name__}.translate")

    def normalize_translation(self, metadata: Dict[str, Any]) -> Dict[str, Any]:
        raise NotImplementedError(f"{self.__class__.__name__}.normalize_translation")


class BaseExtrinsicMapping(BaseMapping):
    """Base class for extrinsic_metadata mappings to inherit from

    To implement a new mapping:

    - inherit this class
    - override translate function
    """

    @classmethod
    def extrinsic_metadata_formats(cls) -> Tuple[str, ...]:
        """
        Returns the list of extrinsic metadata formats which can be translated
        by this mapping
        """
        raise NotImplementedError(f"{cls.__name__}.extrinsic_metadata_formats")

    def normalize_translation(self, metadata: Dict[str, Any]) -> Dict[str, Any]:
        return compact(metadata, forgefed=True)


class BaseIntrinsicMapping(BaseMapping):
    """Base class for intrinsic-metadata mappings to inherit from

    To implement a new mapping:

    - inherit this class
    - override translate function
    """

    @classmethod
    def detect_metadata_files(cls, file_entries: List[DirectoryLsEntry]) -> List[Sha1]:
        """
        Returns the sha1 hashes of files which can be translated by this mapping
        """
        raise NotImplementedError(f"{cls.__name__}.detect_metadata_files")

    def normalize_translation(self, metadata: Dict[str, Any]) -> Dict[str, Any]:
        return compact(metadata, forgefed=False)


class SingleFileIntrinsicMapping(BaseIntrinsicMapping):
    """Base class for all intrinsic metadata mappings that use a single file as input."""

    filename: Union[bytes, Pattern[bytes]]

    @classmethod
    def detect_metadata_files(cls, file_entries: List[DirectoryLsEntry]) -> List[Sha1]:
        filename = cls.filename
        # Check if filename is a regex or bytes:
        if isinstance(filename, bytes):
            for entry in file_entries:
                if entry["name"].lower() == filename.lower():
                    if entry["sha1"] is not None:  # ignore skipped_content and dangling
                        return [entry["sha1"]]
        else:
            for entry in file_entries:
                if filename.match(entry["name"]):
                    if entry["sha1"] is not None:  # ignore skipped_content and dangling
                        return [entry["sha1"]]

        return []


class DictMapping(BaseMapping):
    """Base class for mappings that take as input a file that is mostly
    a key-value store (eg. a shallow JSON dict)."""

    string_fields: List[str] = []
    """List of fields that are simple strings, and don't need any
    normalization."""

    date_fields: List[str] = []
    """List of fields that are strings that should be typed as http://schema.org/Date
    """

    uri_fields: List[str] = []
    """List of fields that are simple URIs, and don't need any
    normalization."""

    @property
    def mapping(self):
        """A translation dict to map dict keys into a canonical name."""
        raise NotImplementedError(f"{self.__class__.__name__}.mapping")

    @staticmethod
    def _normalize_method_name(name: str) -> str:
        return name.replace("-", "_")

    @classmethod
    def supported_terms(cls):
        # one-to-one mapping from the original key to a CodeMeta term
        simple_terms = {
            str(term)
            for (key, term) in cls.mapping.items()
            if key in cls.string_fields + cls.date_fields + cls.uri_fields
            or hasattr(cls, "normalize_" + cls._normalize_method_name(key))
        }

        # more complex mapping from the original key to JSON-LD
        complex_terms = {
            str(term)
            for meth_name in dir(cls)
            if meth_name.startswith("translate_")
            for term in getattr(getattr(cls, meth_name), "produced_terms", [])
        }

        return simple_terms | complex_terms

    def get_root_uri(self, content_dict: Dict) -> rdflib.URIRef:
        """Returns an URI for the SoftwareSourceCode or Repository being described.

        The default implementation uses a temporary URI that is stripped before
        normalization by :meth:`_translate_dict`.
        """
        # The main object being described (the SoftwareSourceCode) does not necessarily
        # may or may not have an id.
        # If it does, it will need to be set by a subclass.
        # If it doesn't we temporarily use this URI to identify it. Unfortunately,
        # we cannot use a blank node as we need to use it for JSON-LD framing later,
        # and blank nodes cannot be used for framing in JSON-LD >= 1.1
        root_id = TMP_ROOT_URI_PREFIX + str(uuid.uuid4())
        return rdflib.URIRef(root_id)

    def _translate_dict(self, content_dict: Dict) -> Dict[str, Any]:
        """
        Translates content  by parsing content from a dict object
        and translating with the appropriate mapping

        Args:
            content_dict (dict): content dict to translate

        Returns:
            dict: translated metadata in json-friendly form needed for
            the indexer

        """
        graph = rdflib.Graph()

        root = self.get_root_uri(content_dict)

        self._translate_to_graph(graph, root, content_dict)

        self.sanitize(graph)

        # Convert from rdflib's internal graph representation to JSON
        s = graph.serialize(format="application/ld+json")

        # Load from JSON to a list of Python objects
        jsonld_graph = json.loads(s)

        # Use JSON-LD framing to turn the graph into a rooted tree
        # frame = {"@type": str(SCHEMA.SoftwareSourceCode)}
        translated_metadata = jsonld.frame(
            jsonld_graph,
            {"@id": str(root)},
            options={
                "documentLoader": _document_loader,
                "processingMode": "json-ld-1.1",
            },
        )

        # Remove the temporary id we added at the beginning
        assert isinstance(translated_metadata["@id"], str)
        if translated_metadata["@id"].startswith(TMP_ROOT_URI_PREFIX):
            del translated_metadata["@id"]

        return self.normalize_translation(translated_metadata)

    def _translate_to_graph(
        self, graph: rdflib.Graph, root: rdflib.term.Identifier, content_dict: Dict
    ) -> None:
        """
        Translates content  by parsing content from a dict object
        and translating with the appropriate mapping to the graph passed as parameter

        Args:
            content_dict (dict): content dict to translate

        """
        graph.add((root, RDF.type, SCHEMA.SoftwareSourceCode))

        for k, v in content_dict.items():
            # First, check if there is a specific translation
            # method for this key
            translation_method = getattr(
                self, "translate_" + self._normalize_method_name(k), None
            )
            if translation_method:
                translation_method(graph, root, v)
            elif k in self.mapping:
                # if there is no method, but the key is known from the
                # crosswalk table
                codemeta_key = self.mapping[k]

                # if there is a normalization method, use it on the value,
                # and add its results to the triples
                normalization_method = getattr(
                    self, "normalize_" + self._normalize_method_name(k), None
                )
                if normalization_method:
                    v = normalization_method(v)
                    if v is None:
                        pass
                    elif isinstance(v, list):
                        for item in reversed(v):
                            if isinstance(item, rdflib.URIRef):
                                add_url_if_valid(graph, root, codemeta_key, str(item))
                            else:
                                graph.add((root, codemeta_key, item))
                    else:
                        if isinstance(v, rdflib.URIRef):
                            add_url_if_valid(graph, root, codemeta_key, str(v))
                        else:
                            graph.add((root, codemeta_key, v))
                elif k in self.string_fields and isinstance(v, str):
                    graph.add((root, codemeta_key, rdflib.Literal(v)))
                elif k in self.string_fields and isinstance(v, list):
                    for item in v:
                        graph.add((root, codemeta_key, rdflib.Literal(item)))
                elif k in self.date_fields and isinstance(v, str):
                    typed_v = rdflib.Literal(v, datatype=SCHEMA.Date)
                    graph.add((root, codemeta_key, typed_v))
                elif k in self.date_fields and isinstance(v, list):
                    for item in v:
                        if isinstance(item, str):
                            typed_item = rdflib.Literal(item, datatype=SCHEMA.Date)
                            graph.add((root, codemeta_key, typed_item))
                elif k in self.uri_fields and isinstance(v, str):
                    add_url_if_valid(graph, root, codemeta_key, v)
                elif k in self.uri_fields and isinstance(v, list):
                    for item in v:
                        add_url_if_valid(graph, root, codemeta_key, item)
                else:
                    continue

        self.extra_translation(graph, root, content_dict)

    def sanitize(self, graph: rdflib.Graph) -> None:
        # Remove triples that make PyLD crash
        for subject, predicate, _ in graph.triples((None, None, rdflib.URIRef(""))):
            graph.remove((subject, predicate, rdflib.URIRef("")))

        # Should not happen, but we's better check as this may lead to incorrect data
        invalid = False
        for triple in graph.triples((rdflib.URIRef(""), None, None)):
            invalid = True
            logging.error("Empty triple subject URI: %r", triple)
        if invalid:
            raise ValueError("Empty triple subject(s)")

    def extra_translation(
        self, graph: rdflib.Graph, root: rdflib.term.Node, d: Dict[str, Any]
    ) -> None:
        """Called at the end of the translation process, and may add arbitrary triples
        to ``graph`` based on the input dictionary (passed as ``d``).
        """
        pass


class JsonMapping(DictMapping):
    """Base class for all mappings that use JSON data as input."""

    def translate(self, raw_content: bytes) -> Optional[Dict]:
        try:
            raw_content_string: str = raw_content.decode()
        except UnicodeDecodeError:
            self.log.warning("Error unidecoding from %s", self.log_suffix)
            return None
        try:
            content_dict = json.loads(raw_content_string)
        except json.JSONDecodeError:
            self.log.warning("Error unjsoning from %s", self.log_suffix)
            return None
        if isinstance(content_dict, dict):
            return self._translate_dict(content_dict)
        return None


class XmlMapping(DictMapping):
    """Base class for all mappings that use XML data as input."""

    def translate(self, raw_content: bytes) -> Optional[Dict]:
        try:
            d = xmltodict.parse(raw_content)
        except xml.parsers.expat.ExpatError:
            self.log.warning("Error parsing XML from %s", self.log_suffix)
            return None
        except UnicodeDecodeError:
            self.log.warning("Error unidecoding XML from %s", self.log_suffix)
            return None
        except (LookupError, ValueError):
            # unknown encoding or multi-byte encoding
            self.log.warning("Error detecting XML encoding from %s", self.log_suffix)
            return None
        if not isinstance(d, dict):
            self.log.warning("Skipping ill-formed XML content: %s", raw_content)
            return None
        return self._translate_dict(d)


class SafeLoader(yaml.SafeLoader):
    yaml_implicit_resolvers = {
        k: [r for r in v if r[0] != "tag:yaml.org,2002:timestamp"]
        for k, v in yaml.SafeLoader.yaml_implicit_resolvers.items()
    }


class YamlMapping(DictMapping):
    """Base class for all mappings that use Yaml data as input."""

    def translate(self, raw_content: bytes) -> Optional[Dict[str, str]]:
        raw_content_string: str = raw_content.decode()
        try:
            content_dict = yaml.load(raw_content_string, Loader=SafeLoader)
        except yaml.YAMLError:
            return None

        if isinstance(content_dict, dict):
            return self._translate_dict(content_dict)

        return None



# Copyright (C) 2018-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import email.parser
import email.policy

from rdflib import BNode, Literal, URIRef

from swh.indexer.codemeta import CROSSWALK_TABLE
from swh.indexer.namespaces import RDF, SCHEMA

from .base import DictMapping, SingleFileIntrinsicMapping
from .utils import add_list

_normalize_pkginfo_key = str.lower


class LinebreakPreservingEmailPolicy(email.policy.EmailPolicy):
    def header_fetch_parse(self, name, value):
        if hasattr(value, "name"):
            return value
        value = value.replace("\n        ", "\n")
        return self.header_factory(name, value)


class PythonPkginfoMapping(DictMapping, SingleFileIntrinsicMapping):
    """Dedicated class for Python's PKG-INFO mapping and translation.

    https://www.python.org/dev/peps/pep-0314/"""

    name = "pkg-info"
    filename = b"PKG-INFO"
    mapping = {
        _normalize_pkginfo_key(k): v
        for (k, v) in CROSSWALK_TABLE["Python PKG-INFO"].items()
    }
    string_fields = [
        "name",
        "version",
        "description",
        "summary",
        "author",
        "author-email",
    ]

    _parser = email.parser.BytesHeaderParser(policy=LinebreakPreservingEmailPolicy())

    def translate(self, content):
        msg = self._parser.parsebytes(content)
        d = {}
        for key, value in msg.items():
            key = _normalize_pkginfo_key(key)
            if value != "UNKNOWN":
                d.setdefault(key, []).append(value)
        return self._translate_dict(d)

    def extra_translation(self, graph, root, d):
        author_names = list(graph.triples((root, SCHEMA.author, None)))
        author_emails = list(graph.triples((root, SCHEMA.email, None)))
        graph.remove((root, SCHEMA.author, None))
        graph.remove((root, SCHEMA.email, None))
        if author_names or author_emails:
            author = BNode()
            graph.add((author, RDF.type, SCHEMA.Person))
            for _, _, author_name in author_names:
                graph.add((author, SCHEMA.name, author_name))
            for _, _, author_email in author_emails:
                graph.add((author, SCHEMA.email, author_email))
            add_list(graph, root, SCHEMA.author, [author])

    def normalize_home_page(self, urls):
        return [URIRef(url) for url in urls]

    def normalize_keywords(self, keywords):
        return [Literal(keyword) for s in keywords for keyword in s.split(" ")]

    def normalize_license(self, licenses):
        return [URIRef("https://spdx.org/licenses/" + license) for license in licenses]



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.indexer.metadata_dictionary import MAPPINGS


def test_compute_metadata_composer():
    raw_content = """{
"name": "symfony/polyfill-mbstring",
"type": "library",
"description": "Symfony polyfill for the Mbstring extension",
"keywords": [
    "polyfill",
    "shim",
    "compatibility",
    "portable"
],
"homepage": "https://symfony.com",
"license": "MIT",
"authors": [
    {
        "name": "Nicolas Grekas",
        "email": "p@tchwork.com"
    },
    {
        "name": "Symfony Community",
        "homepage": "https://symfony.com/contributors"
    }
],
"require": {
    "php": ">=7.1"
},
"provide": {
    "ext-mbstring": "*"
},
"autoload": {
    "files": [
        "bootstrap.php"
    ]
},
"suggest": {
    "ext-mbstring": "For best performance"
},
"minimum-stability": "dev",
"extra": {
    "branch-alias": {
        "dev-main": "1.26-dev"
    },
    "thanks": {
        "name": "symfony/polyfill",
        "url": "https://github.com/symfony/polyfill"
    }
}
}
    """.encode(
        "utf-8"
    )

    result = MAPPINGS["ComposerMapping"]().translate(raw_content)

    assert set(result.pop("keywords")) == {
        "polyfill",
        "shim",
        "compatibility",
        "portable",
    }, result
    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "symfony/polyfill-mbstring",
        "description": "Symfony polyfill for the Mbstring extension",
        "url": "https://symfony.com",
        "license": "https://spdx.org/licenses/MIT",
        "author": [
            {
                "type": "Person",
                "name": "Nicolas Grekas",
                "email": "p@tchwork.com",
            },
            {
                "type": "Person",
                "name": "Symfony Community",
            },
        ],
    }

    assert result == expected



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.indexer.metadata_dictionary import MAPPINGS


def test_compute_metadata_pubspec():
    raw_content = b"""
---
name: newtify
description: >-
  Have you been turned into a newt?  Would you like to be?
  This package can help. It has all of the
  newt-transmogrification functionality you have been looking
  for.
keywords:
  - polyfill
  - shim
  - compatibility
  - portable
  - mbstring
version: 1.2.3
license: MIT
homepage: https://example-pet-store.com/newtify
documentation: https://example-pet-store.com/newtify/docs

environment:
  sdk: '>=2.10.0 <3.0.0'

dependencies:
  efts: ^2.0.4
  transmogrify: ^0.4.0

dev_dependencies:
  test: '>=1.15.0 <2.0.0'
    """

    result = MAPPINGS["PubMapping"]().translate(raw_content)

    assert set(result.pop("keywords")) == {
        "polyfill",
        "shim",
        "compatibility",
        "portable",
        "mbstring",
    }, result
    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "newtify",
        "description": """Have you been turned into a newt?  Would you like to be? \
This package can help. It has all of the \
newt-transmogrification functionality you have been looking \
for.""",
        "url": "https://example-pet-store.com/newtify",
        "license": "https://spdx.org/licenses/MIT",
    }

    assert result == expected


def test_normalize_author_pubspec():
    raw_content = b"""
    author: Atlee Pine <atlee@example.org>
    """

    result = MAPPINGS["PubMapping"]().translate(raw_content)

    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [
            {"type": "Person", "name": "Atlee Pine", "email": "atlee@example.org"},
        ],
    }

    assert result == expected


def test_normalize_authors_pubspec():
    raw_content = b"""
    authors:
      - Vicky Merzown <vmz@example.org>
      - Ron Bilius Weasley
    """

    result = MAPPINGS["PubMapping"]().translate(raw_content)

    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [
            {"type": "Person", "name": "Vicky Merzown", "email": "vmz@example.org"},
            {
                "type": "Person",
                "name": "Ron Bilius Weasley",
            },
        ],
    }

    assert result == expected


@pytest.mark.xfail(reason="https://github.com/w3c/json-ld-api/issues/547")
def test_normalize_author_authors_pubspec():
    raw_content = b"""
    authors:
      - Vicky Merzown <vmz@example.org>
      - Ron Bilius Weasley
    author: Hermione Granger
    """

    result = MAPPINGS["PubMapping"]().translate(raw_content)

    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [
            {"type": "Person", "name": "Vicky Merzown", "email": "vmz@example.org"},
            {
                "type": "Person",
                "name": "Ron Bilius Weasley",
            },
            {
                "type": "Person",
                "name": "Hermione Granger",
            },
        ],
    }

    assert result == expected


def test_normalize_empty_authors():
    raw_content = b"""
    authors:
    """

    result = MAPPINGS["PubMapping"]().translate(raw_content)

    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    assert result == expected


def test_invalid_yaml():
    raw_content = b"""
    name: smartech_push
    license: { :type => "Commercial", :file => "LICENSE" }
    """

    result = MAPPINGS["PubMapping"]().translate(raw_content)

    assert result is None


def test_invalid_tag():
    raw_content = b"""
    name: translatron
    description: !BETA VERSION - NOT FOR LIVE OR PROD USAGE!
    """

    result = MAPPINGS["PubMapping"]().translate(raw_content)

    assert result is None



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict
import itertools
import logging
from typing import Any, Dict, Generator, List, Optional, Tuple, cast

import sentry_sdk

from swh.core import utils
from swh.core.config import load_from_envvar
from swh.model import hashutil
from swh.model.model import Content
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.factory import get_objstorage
from swh.objstorage.interface import ObjId
from swh.storage import get_storage

DEFAULT_CONFIG: Dict[str, Any] = {
    "storage": {"cls": "memory"},
    "objstorage": {"cls": "memory"},
    # the set of checksums that should be computed.
    # Examples: 'sha1_git', 'blake2b512', 'blake2s256'
    "compute_checksums": [],
    # whether checksums that already exist in the DB should be
    # recomputed/updated or left untouched
    "recompute_checksums": False,
    # Number of contents to retrieve blobs at the same time
    "batch_size_retrieve_content": 10,
    # Number of contents to update at the same time
    "batch_size_update": 100,
}


class RecomputeChecksums:
    """Class in charge of (re)computing content's hashes.

    Hashes to compute are defined across 2 configuration options:

    compute_checksums ([str])
      list of hash algorithms that
      py:func:`swh.model.hashutil.MultiHash.from_data` function should
      be able to deal with. For variable-length checksums, a desired
      checksum length should also be provided. Their format is
      <algorithm's name>:<variable-length> e.g: blake2:512

    recompute_checksums (bool)
      a boolean to notify that we also want to recompute potential existing
      hashes specified in compute_checksums.  Default to False.

    """

    def __init__(self) -> None:
        self.config = load_from_envvar(DEFAULT_CONFIG)
        self.storage = get_storage(**self.config["storage"])
        self.objstorage = get_objstorage(**self.config["objstorage"])
        self.compute_checksums = self.config["compute_checksums"]
        self.recompute_checksums = self.config["recompute_checksums"]
        self.batch_size_retrieve_content = self.config["batch_size_retrieve_content"]
        self.batch_size_update = self.config["batch_size_update"]
        self.log = logging.getLogger("swh.indexer.rehash")

        if not self.compute_checksums:
            raise ValueError("Checksums list should not be empty.")

    def _read_content_ids(
        self, contents: List[Dict[str, Any]]
    ) -> Generator[bytes, Any, None]:
        """Read the content identifiers from the contents."""
        for c in contents:
            h = c["sha1"]
            if isinstance(h, str):
                h = hashutil.hash_to_bytes(h)

            yield h

    def get_new_contents_metadata(
        self, all_contents: List[Dict[str, Any]]
    ) -> Generator[Tuple[Dict[str, Any], List[Any]], Any, None]:
        """Retrieve raw contents and compute new checksums on the
           contents. Unknown or corrupted contents are skipped.

        Args:
            all_contents: List of contents as dictionary with
              the necessary primary keys

        Yields:
            tuple: tuple of (content to update, list of checksums computed)

        """
        content_ids = self._read_content_ids(all_contents)
        for contents in utils.grouper(content_ids, self.batch_size_retrieve_content):
            contents_iter = itertools.tee(contents, 2)
            try:
                sha1s = [s for s in contents_iter[0]]
                content_metadata: List[Optional[Content]] = self.storage.content_get(
                    sha1s
                )
            except Exception:
                self.log.exception("Problem when reading contents metadata.")
                sentry_sdk.capture_exception()
                continue

            for content in content_metadata:
                if not content:
                    continue
                # Recompute checksums provided in compute_checksums options
                if self.recompute_checksums:
                    checksums_to_compute = list(self.compute_checksums)
                else:
                    # Compute checksums provided in compute_checksums
                    # options not already defined for that content
                    checksums_to_compute = [
                        h for h in self.compute_checksums if not content.get_hash(h)
                    ]

                if not checksums_to_compute:  # Nothing to recompute
                    continue

                try:
                    raw_content = self.objstorage.get(cast(ObjId, content.hashes()))
                except ObjNotFoundError:
                    self.log.warning(
                        "Content %s not found in objstorage!", content.hashes()
                    )
                    continue

                content_hashes = hashutil.MultiHash.from_data(
                    raw_content, hash_names=checksums_to_compute
                ).digest()
                content_dict = content.to_dict()
                content_dict.update(content_hashes)
                yield content_dict, checksums_to_compute

    def run(self, contents: List[Dict[str, Any]]) -> Dict:
        """Given a list of content:

          - (re)compute a given set of checksums on contents available in our
            object storage
          - update those contents with the new metadata

        Args:
            contents: contents as dictionary with necessary keys.
                key present in such dictionary should be the ones defined in
                the 'primary_key' option.

        Returns:
            A summary dict with key 'status', task' status and 'count' the
            number of updated contents.

        """
        status = "uneventful"
        count = 0
        for data in utils.grouper(
            self.get_new_contents_metadata(contents), self.batch_size_update
        ):
            groups: Dict[str, List[Any]] = defaultdict(list)
            for content, keys_to_update in data:
                keys_str = ",".join(keys_to_update)
                groups[keys_str].append(content)

            for keys_to_update, contents in groups.items():
                keys: List[str] = keys_to_update.split(",")
                try:
                    self.storage.content_update(contents, keys=keys)
                    count += len(contents)
                    status = "eventful"
                except Exception:
                    self.log.exception("Problem during update.")
                    sentry_sdk.capture_exception()
                    continue

        return {
            "status": status,
            "count": count,
        }



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import List
import urllib.parse

from rdflib import BNode, Graph, Literal, URIRef
import rdflib.term

from swh.indexer.codemeta import CROSSWALK_TABLE
from swh.indexer.namespaces import RDF, SCHEMA

from .base import SingleFileIntrinsicMapping, YamlMapping
from .utils import add_map

DOI = URIRef("https://doi.org/")
SPDX = URIRef("https://spdx.org/licenses/")


class CffMapping(YamlMapping, SingleFileIntrinsicMapping):
    """Dedicated class for Citation (CITATION.cff) mapping and translation"""

    name = "cff"
    filename = b"CITATION.cff"
    mapping = CROSSWALK_TABLE["Citation File Format Core (CFF-Core) 1.0.2"]
    string_fields = ["title", "keywords", "license", "abstract", "version", "doi"]
    date_fields = ["date-released"]
    uri_fields = ["url", "repository-code"]

    def _translate_author(self, graph: Graph, author: dict) -> rdflib.term.Node:
        node: rdflib.term.Node
        if (
            "orcid" in author
            and isinstance(author["orcid"], str)
            and urllib.parse.urlparse(author["orcid"]).netloc
        ):
            node = URIRef(author["orcid"])
        else:
            node = BNode()
        graph.add((node, RDF.type, SCHEMA.Person))
        if "affiliation" in author and isinstance(author["affiliation"], str):
            affiliation = BNode()
            graph.add((node, SCHEMA.affiliation, affiliation))
            graph.add((affiliation, RDF.type, SCHEMA.Organization))
            graph.add((affiliation, SCHEMA.name, Literal(author["affiliation"])))
        if "family-names" in author and isinstance(author["family-names"], str):
            graph.add((node, SCHEMA.familyName, Literal(author["family-names"])))
        if "given-names" in author and isinstance(author["given-names"], str):
            graph.add((node, SCHEMA.givenName, Literal(author["given-names"])))
        return node

    def translate_authors(
        self, graph: Graph, root: URIRef, authors: List[dict]
    ) -> None:
        add_map(graph, root, SCHEMA.author, self._translate_author, authors)

    def normalize_doi(self, s: str) -> URIRef:
        if isinstance(s, str):
            return DOI + s

    def normalize_license(self, s: str) -> URIRef:
        if isinstance(s, str):
            return SPDX + s



# Copyright (C) 2016-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


# implemented as a function to help lazy loading
def get_datastore(*args, **kw):
    from .indexer import get_indexer_storage

    return get_indexer_storage(*args, **kw)


default_cfg = {
    "default_interval": "1 day",
    "min_interval": "12 hours",
    "max_interval": "1 day",
    "backoff_factor": 2,
    "max_queue_length": 5000,
}



# Copyright (C) 2018-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import json
import logging
import re
from typing import Any, Dict, List, Optional, Tuple, Union
import xml.etree.ElementTree as ET

import iso8601
import xmltodict

from swh.indexer.codemeta import CODEMETA_CONTEXT_URL, CODEMETA_TERMS, compact, expand

from .base import BaseExtrinsicMapping, SingleFileIntrinsicMapping

ATOM_URI = "http://www.w3.org/2005/Atom"

_TAG_RE = re.compile(r"\{(?P<namespace>.*?)\}(?P<localname>.*)")
_IGNORED_NAMESPACES = ("http://www.w3.org/2005/Atom",)
_DATE_RE = re.compile("^[0-9]{4}-[0-9]{1,2}-[0-9]{1,2}$")

logger = logging.getLogger(__name__)


class CodemetaMapping(SingleFileIntrinsicMapping):
    """
    dedicated class for CodeMeta (codemeta.json) mapping and translation
    """

    name = "codemeta"
    filename = b"codemeta.json"
    string_fields = None

    @classmethod
    def supported_terms(cls) -> List[str]:
        return [term for term in CODEMETA_TERMS if not term.startswith("@")]

    def translate(self, content: bytes) -> Optional[Dict[str, Any]]:
        try:
            return self.normalize_translation(expand(json.loads(content.decode())))
        except Exception:
            return None


class SwordCodemetaMapping(BaseExtrinsicMapping):
    """
    dedicated class for mapping and translation from JSON-LD statements
    embedded in SWORD documents, optionally using Codemeta contexts,
    as described in the :ref:`deposit-protocol`.
    """

    name = "sword-codemeta"

    @classmethod
    def extrinsic_metadata_formats(cls) -> Tuple[str, ...]:
        return (
            "sword-v2-atom-codemeta",
            "sword-v2-atom-codemeta-v2",
        )

    @classmethod
    def supported_terms(cls) -> List[str]:
        return [term for term in CODEMETA_TERMS if not term.startswith("@")]

    def xml_to_jsonld(self, e: ET.Element) -> Union[str, Dict[str, Any]]:
        # Keys are JSON-LD property names (URIs or terms).
        # Values are either a single string (if key is "type") or list of
        # other dicts with the same type recursively.
        # To simply annotations, we omit the single string case here.
        doc: Dict[str, List[Union[str, Dict[str, Any]]]] = collections.defaultdict(list)

        for child in e:
            m = _TAG_RE.match(child.tag)
            assert m, f"Tag with no namespace: {child}"
            namespace = m.group("namespace")
            localname = m.group("localname")
            if namespace == ATOM_URI and localname in ("title", "name"):
                # Convert Atom to Codemeta name; in case codemeta:name
                # is not provided or different
                doc["name"].append(self.xml_to_jsonld(child))
            elif namespace == ATOM_URI and localname in ("author", "email"):
                # ditto for these author properties (note that author email is also
                # covered by the previous test)
                doc[localname].append(self.xml_to_jsonld(child))
            elif namespace in _IGNORED_NAMESPACES:
                # SWORD-specific namespace that is not interesting to translate
                pass
            elif namespace.lower() == CODEMETA_CONTEXT_URL:
                # It is a term defined by the context; write is as-is and JSON-LD
                # expansion will convert it to a full URI based on
                # "@context": CODEMETA_CONTEXT_URL
                jsonld_child = self.xml_to_jsonld(child)
                if (
                    localname
                    in (
                        "dateCreated",
                        "dateModified",
                        "datePublished",
                    )
                    and isinstance(jsonld_child, str)
                    and _DATE_RE.match(jsonld_child)
                ):
                    # Dates missing a leading zero for their day/month, used
                    # to be allowed by the deposit; so we need to reformat them
                    # to be valid ISO8601.
                    jsonld_child = iso8601.parse_date(jsonld_child).date().isoformat()
                if localname == "id":
                    # JSON-LD only allows a single id, and they have to be strings.
                    if localname in doc:
                        logger.error(
                            "Duplicate <id>s in SWORD document: %r and %r",
                            doc[localname],
                            jsonld_child,
                        )
                        continue
                    elif not jsonld_child:
                        logger.error("Empty <id> value in SWORD document")
                        continue
                    elif not isinstance(jsonld_child, str):
                        logger.error(
                            "Unexpected <id> value in SWORD document: %r", jsonld_child
                        )
                        continue
                    else:
                        doc[localname] = jsonld_child  # type: ignore[assignment]
                else:
                    doc[localname].append(jsonld_child)
            else:
                # Otherwise, we already know the URI
                doc[f"{namespace}{localname}"].append(self.xml_to_jsonld(child))

        # The above needed doc values to be list to work; now we allow any type
        # of value as key "@value" cannot have a list as value.
        doc_: Dict[str, Any] = doc

        text = e.text.strip() if e.text else None
        if text:
            # TODO: check doc is empty, and raise mixed-content error otherwise?
            return text

        return doc_

    def translate(self, content: bytes) -> Optional[Dict[str, Any]]:
        # Parse XML
        try:
            root = ET.fromstring(content)
        except ET.ParseError:
            logger.error("Failed to parse XML document: %s", content)
            return None
        else:
            # Transform to JSON-LD document
            doc = self.xml_to_jsonld(root)

            assert isinstance(doc, dict), f"Root object is not a dict: {doc}"

            # Add @context to JSON-LD expansion replaces the "codemeta:" prefix
            # hash (which uses the context URL as namespace URI for historical
            # reasons) into properties in `http://schema.org/` and
            # `https://codemeta.github.io/terms/` namespaces
            doc["@context"] = CODEMETA_CONTEXT_URL

            # Normalize as a Codemeta document
            return self.normalize_translation(expand(doc))

    def normalize_translation(self, metadata: Dict[str, Any]) -> Dict[str, Any]:
        return compact(metadata, forgefed=False)


class JsonSwordCodemetaMapping(SwordCodemetaMapping):
    """
    Variant of :class:`SwordCodemetaMapping` that reads the legacy
    ``sword-v2-atom-codemeta-v2-in-json`` format and converts it back to
    ``sword-v2-atom-codemeta-v2`` XML
    """

    name = "json-sword-codemeta"

    @classmethod
    def extrinsic_metadata_formats(cls) -> Tuple[str, ...]:
        return ("sword-v2-atom-codemeta-v2-in-json",)

    def translate(self, content: bytes) -> Optional[Dict[str, Any]]:
        # ``content`` was generated by calling ``xmltodict.parse()`` on a XML document,
        # so ``xmltodict.unparse()`` is guaranteed to return a document that is
        # semantically equivalent to the original and pass it to SwordCodemetaMapping.
        try:
            json_doc = json.loads(content)
        except json.JSONDecodeError:
            logger.error("Failed to parse JSON document: %s", content)
            return None
        else:
            if json_doc.get("@xmlns") != ATOM_URI:
                # Technically, non-default XMLNS were allowed, but it does not seem like
                # anyone used them, so they do not need to be implemented here.
                raise NotImplementedError(f"Unexpected XMLNS set: {json_doc}")

            # Root tag was stripped by swh-deposit
            json_doc = {"entry": json_doc}

            return super().translate(xmltodict.unparse(json_doc))



# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re
from typing import List, Optional, Tuple, Union

from swh.model.model import Snapshot, SnapshotBranch, SnapshotTargetType
from swh.model.swhids import CoreSWHID, ObjectType
from swh.storage.algos.origin import origin_get_latest_visit_status
from swh.storage.algos.snapshot import snapshot_get_all_branches
from swh.storage.interface import PartialBranches, StorageInterface


def get_head_swhid(storage: StorageInterface, origin_url: str) -> Optional[CoreSWHID]:
    """Returns the SWHID of the head revision or release of an origin"""
    visit_status = origin_get_latest_visit_status(
        storage, origin_url, allowed_statuses=["full"], require_snapshot=True
    )
    if not visit_status:
        return None
    assert visit_status.snapshot is not None

    if visit_status.type == "ftp":
        # We need to fetch all branches in order to find the largest one
        snapshot = snapshot_get_all_branches(storage, visit_status.snapshot)
        if snapshot is None:
            return None
        return _try_get_ftp_head(storage, snapshot)
    else:
        # Peak into the snapshot, without fetching too many refs.
        # If the snapshot is small, this gets all of it in a single request.
        # If the snapshot is large, we will query specific branches as we need them.
        partial_branches = storage.snapshot_get_branches(
            visit_status.snapshot, branches_count=100
        )
        if partial_branches is None:
            # Snapshot does not exist
            return None
        return _try_get_head_generic(storage, partial_branches)


_archive_filename_re = re.compile(
    rb"^"
    rb"(?P<pkgname>.*)[-_]"
    rb"(?P<version>[0-9]+(\.[0-9])*)"
    rb"(?P<preversion>[-+][a-zA-Z0-9.~]+?)?"
    rb"(?P<extension>(\.[a-zA-Z0-9]+)+)"
    rb"$"
)


def _parse_version(filename: bytes) -> Tuple[Union[float, int, str], ...]:
    """Extracts the release version from an archive filename,
    to get an ordering whose maximum is likely to be the last
    version of the software

    >>> _parse_version(b'foo')
    (-inf,)
    >>> _parse_version(b'foo.tar.gz')
    (-inf,)
    >>> _parse_version(b'gnu-hello-0.0.1.tar.gz')
    (0, 0, 1, 0)
    >>> _parse_version(b'gnu-hello-0.0.1-beta2.tar.gz')
    (0, 0, 1, -1, 'beta2')
    >>> _parse_version(b'gnu-hello-0.0.1+foobar.tar.gz')
    (0, 0, 1, 1, 'foobar')
    """
    res = _archive_filename_re.match(filename)
    if res is None:
        return (float("-infinity"),)
    version: List[Union[float, int, str]] = [
        int(n) for n in res.group("version").decode().split(".")
    ]
    if res.group("preversion") is None:
        version.append(0)
    else:
        preversion = res.group("preversion").decode()
        if preversion.startswith("-"):
            version.append(-1)
            version.append(preversion[1:])
        elif preversion.startswith("+"):
            version.append(1)
            version.append(preversion[1:])
        else:
            assert False, res.group("preversion")
    return tuple(version)


def _try_get_ftp_head(
    storage: StorageInterface, snapshot: Snapshot
) -> Optional[CoreSWHID]:
    archive_names = list(snapshot.branches)
    max_archive_name = max(archive_names, key=_parse_version)
    return _try_resolve_target(
        storage,
        {"id": snapshot.id, "branches": dict(snapshot.branches), "next_branch": None},
        branch_name=max_archive_name,
    )


def _try_get_head_generic(
    storage: StorageInterface, partial_branches: PartialBranches
) -> Optional[CoreSWHID]:
    # Works on 'deposit', 'pypi', and VCSs.
    return _try_resolve_target(
        storage, partial_branches, branch_name=b"HEAD"
    ) or _try_resolve_target(storage, partial_branches, branch_name=b"master")


def _get_branch(
    storage: StorageInterface, partial_branches: PartialBranches, branch_name: bytes
) -> Optional[SnapshotBranch]:
    """Given a ``branch_name``, gets it from ``partial_branches`` if present,
    and fetches it from the storage otherwise."""
    if branch_name in partial_branches["branches"]:
        return partial_branches["branches"][branch_name]
    elif partial_branches["next_branch"] is not None:
        # Branch is not in `partial_branches`, and `partial_branches` indeed partial
        res = storage.snapshot_get_branches(
            partial_branches["id"], branches_from=branch_name, branches_count=1
        )
        assert res is not None, "Snapshot does not exist anymore"
        return res["branches"].get(branch_name)
    else:
        # Branch is not in `partial_branches`, but `partial_branches` is the full
        # list of branches, which means it is a dangling reference.
        return None


def _try_resolve_target(
    storage: StorageInterface, partial_branches: PartialBranches, branch_name: bytes
) -> Optional[CoreSWHID]:
    try:
        branch = _get_branch(storage, partial_branches, branch_name)
        if branch is None:
            return None

        while branch.target_type == SnapshotTargetType.ALIAS:
            branch = _get_branch(storage, partial_branches, branch.target)
            if branch is None:
                return None

        if branch.target_type == SnapshotTargetType.REVISION:
            return CoreSWHID(object_type=ObjectType.REVISION, object_id=branch.target)
        elif branch.target_type == SnapshotTargetType.CONTENT:
            return None  # TODO
        elif branch.target_type == SnapshotTargetType.DIRECTORY:
            return None  # TODO
        elif branch.target_type == SnapshotTargetType.RELEASE:
            return CoreSWHID(object_type=ObjectType.RELEASE, object_id=branch.target)
        else:
            assert False, branch
    except KeyError:
        return None



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from hypothesis import HealthCheck, given, settings
import pytest

from swh.indexer.metadata_detector import detect_metadata
from swh.indexer.metadata_dictionary import MAPPINGS
from swh.indexer.storage.model import ContentMetadataRow

from ..test_metadata import TRANSLATOR_TOOL, ContentMetadataTestIndexer
from ..utils import (
    BASE_TEST_CONFIG,
    MAPPING_DESCRIPTION_CONTENT_SHA1,
    json_document_strategy,
)


def test_compute_metadata_none():
    """
    testing content empty content is empty
    should return None
    """
    content = b""

    # None if no metadata was found or an error occurred
    declared_metadata = None
    result = MAPPINGS["NpmMapping"]().translate(content)
    assert declared_metadata == result


def test_compute_metadata_npm():
    """
    testing only computation of metadata with hard_mapping_npm
    """
    content = b"""
        {
            "name": "test_metadata",
            "version": "0.0.2",
            "description": "Simple package.json test for indexer",
              "repository": {
                "type": "git",
                "url": "https://github.com/moranegg/metadata_test"
            },
            "author": {
                "email": "moranegg@example.com",
                "name": "Morane G"
            }
        }
    """
    declared_metadata = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "test_metadata",
        "version": "0.0.2",
        "description": "Simple package.json test for indexer",
        "codeRepository": "git+https://github.com/moranegg/metadata_test",
        "author": [
            {
                "type": "Person",
                "name": "Morane G",
                "email": "moranegg@example.com",
            }
        ],
    }

    result = MAPPINGS["NpmMapping"]().translate(content)
    assert declared_metadata == result


def test_compute_metadata_invalid_description_npm():
    """
    testing only computation of metadata with hard_mapping_npm
    """
    content = b"""
        {
            "name": "test_metadata",
            "version": "0.0.2",
            "description": 1234
    }
    """
    declared_metadata = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "test_metadata",
        "version": "0.0.2",
    }

    result = MAPPINGS["NpmMapping"]().translate(content)
    assert declared_metadata == result


def test_index_content_metadata_npm(storage, obj_storage):
    """
    testing NPM with package.json
    - one sha1 uses a file that can't be translated to metadata and
      should return None in the translated metadata
    """
    sha1s = [
        MAPPING_DESCRIPTION_CONTENT_SHA1["json:test-metadata-package.json"],
        MAPPING_DESCRIPTION_CONTENT_SHA1["json:npm-package.json"],
        MAPPING_DESCRIPTION_CONTENT_SHA1["python:code"],
    ]

    # this metadata indexer computes only metadata for package.json
    # in npm context with a hard mapping
    config = BASE_TEST_CONFIG.copy()
    config["tools"] = [TRANSLATOR_TOOL]
    metadata_indexer = ContentMetadataTestIndexer(config=config)
    metadata_indexer.run(sha1s, log_suffix="unknown content")
    results = list(metadata_indexer.idx_storage.content_metadata_get(sha1s))

    expected_results = [
        ContentMetadataRow(
            id=sha1s[0],
            tool=TRANSLATOR_TOOL,
            metadata={
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "codeRepository": "git+https://github.com/moranegg/metadata_test",
                "description": "Simple package.json test for indexer",
                "name": "test_metadata",
                "version": "0.0.1",
            },
        ),
        ContentMetadataRow(
            id=sha1s[1],
            tool=TRANSLATOR_TOOL,
            metadata={
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "issueTracker": "https://github.com/npm/npm/issues",
                "author": [
                    {
                        "type": "Person",
                        "name": "Isaac Z. Schlueter",
                        "email": "i@izs.me",
                        "url": "http://blog.izs.me",
                    }
                ],
                "codeRepository": "git+https://github.com/npm/npm",
                "description": "a package manager for JavaScript",
                "license": "https://spdx.org/licenses/Artistic-2.0",
                "version": "5.0.3",
                "name": "npm",
                "url": "https://docs.npmjs.com/",
            },
        ),
    ]

    for result in results:
        del result.tool["id"]
        result.metadata.pop("keywords", None)

    # The assertion below returns False sometimes because of nested lists
    assert expected_results == results


def test_npm_null_list_item_normalization():
    package_json = b"""{
        "name": "foo",
        "keywords": [
            "foo",
            null
        ],
        "homepage": [
            "http://example.org/",
            null
        ]
    }"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "name": "foo",
        "type": "SoftwareSourceCode",
        "url": "http://example.org/",
        "keywords": "foo",
    }


def test_npm_bugs_normalization():
    # valid dictionary
    package_json = b"""{
        "name": "foo",
        "bugs": {
            "url": "https://github.com/owner/project/issues",
            "email": "foo@example.com"
        }
    }"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "name": "foo",
        "issueTracker": "https://github.com/owner/project/issues",
        "type": "SoftwareSourceCode",
    }

    # "invalid" dictionary
    package_json = b"""{
        "name": "foo",
        "bugs": {
            "email": "foo@example.com"
        }
    }"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "name": "foo",
        "type": "SoftwareSourceCode",
    }

    # string
    package_json = b"""{
        "name": "foo",
        "bugs": "https://github.com/owner/project/issues"
    }"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "name": "foo",
        "issueTracker": "https://github.com/owner/project/issues",
        "type": "SoftwareSourceCode",
    }


def test_npm_repository_normalization():
    # normal
    package_json = b"""{
        "name": "foo",
        "repository": {
            "type" : "git",
            "url" : "https://github.com/npm/cli.git"
        }
    }"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "name": "foo",
        "codeRepository": "git+https://github.com/npm/cli.git",
        "type": "SoftwareSourceCode",
    }

    # missing url
    package_json = b"""{
        "name": "foo",
        "repository": {
            "type" : "git"
        }
    }"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "name": "foo",
        "type": "SoftwareSourceCode",
    }

    # github shortcut
    package_json = b"""{
        "name": "foo",
        "repository": "github:npm/cli"
    }"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    expected_result = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "name": "foo",
        "codeRepository": "git+https://github.com/npm/cli.git",
        "type": "SoftwareSourceCode",
    }
    assert result == expected_result

    # github shortshortcut
    package_json = b"""{
        "name": "foo",
        "repository": "npm/cli"
    }"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == expected_result

    # gitlab shortcut
    package_json = b"""{
        "name": "foo",
        "repository": "gitlab:user/repo"
    }"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "name": "foo",
        "codeRepository": "git+https://gitlab.com/user/repo.git",
        "type": "SoftwareSourceCode",
    }


def test_npm_author():
    package_json = rb"""{
  "version": "1.0.0",
  "author": "Foo Bar (@example)"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [{"name": "Foo Bar", "type": "Person"}],
        "version": "1.0.0",
    }


def test_npm_invalid_uris():
    package_json = rb"""{
  "version": "1.0.0",
  "homepage": "",
  "author": {
    "name": "foo",
    "url": "http://example.org"
  }
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [{"name": "foo", "type": "Person", "url": "http://example.org"}],
        "version": "1.0.0",
    }

    package_json = rb"""{
  "version": "1.0.0",
  "homepage": "http://example.org",
  "author": {
    "name": "foo",
    "url": ""
  }
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [{"name": "foo", "type": "Person"}],
        "url": "http://example.org",
        "version": "1.0.0",
    }

    package_json = rb"""{
  "version": "1.0.0",
  "homepage": "",
  "author": {
    "name": "foo",
    "url": ""
  },
  "bugs": ""
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [{"name": "foo", "type": "Person"}],
        "version": "1.0.0",
    }

    package_json = rb"""{
  "version": "1.0.0",
  "homepage": "http:example.org",
  "author": {
    "name": "foo",
    "url": "http:example.com"
  },
  "bugs": {
    "url": "http:example.com"
  }
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [{"name": "foo", "type": "Person"}],
        "version": "1.0.0",
    }

    package_json = rb"""{
  "version": "1.0.0",
  "repository": "git+https://g ithub.com/foo/bar.git"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "version": "1.0.0",
    }

    package_json = rb"""{
  "version": "1.0.0",
  "repository": "git+http://\\u001b[D\\u001b[D\\u001b[Ds\\u001b[C\\u001b[C\\u001b[D\\u001b://github.com/dearzoe/array-combination"
}"""  # noqa
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "version": "1.0.0",
    }


def test_npm_invalid_licenses():
    package_json = rb"""{
  "version": "1.0.0",
  "license": "SEE LICENSE IN LICENSE.md",
  "author": {
    "name": "foo",
    "url": "http://example.org"
  }
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [{"name": "foo", "type": "Person", "url": "http://example.org"}],
        "version": "1.0.0",
    }


@settings(suppress_health_check=[HealthCheck.too_slow])
@given(json_document_strategy(keys=list(MAPPINGS["NpmMapping"].mapping)))  # type: ignore
def test_npm_adversarial(doc):
    raw = json.dumps(doc).encode()
    MAPPINGS["NpmMapping"]().translate(raw)


@pytest.mark.parametrize(
    "filename", [b"package.json", b"Package.json", b"PACKAGE.json", b"PACKAGE.JSON"]
)
def test_detect_metadata_package_json(filename):
    df = [
        {
            "sha1_git": b"abc",
            "name": b"index.js",
            "target": b"abc",
            "length": 897,
            "status": "visible",
            "type": "file",
            "perms": 33188,
            "dir_id": b"dir_a",
            "sha1": b"bcd",
        },
        {
            "sha1_git": b"aab",
            "name": filename,
            "target": b"aab",
            "length": 712,
            "status": "visible",
            "type": "file",
            "perms": 33188,
            "dir_id": b"dir_a",
            "sha1": b"cde",
        },
    ]
    results = detect_metadata(df)

    expected_results = {"NpmMapping": [b"cde"]}
    assert expected_results == results


def test_valid_spdx_expressions():
    package_json = rb"""{
  "license": "Apache-2.0"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "license": "https://spdx.org/licenses/Apache-2.0",
    }

    package_json = rb"""{
  "license": "(Apache-2.0)"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "license": "https://spdx.org/licenses/Apache-2.0",
    }

    package_json = rb"""{
  "license": "MIT OR Apache-2.0"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "license": [
            "https://spdx.org/licenses/Apache-2.0",
            "https://spdx.org/licenses/MIT",
        ],
    }

    package_json = rb"""{
  "license": "(MIT OR Apache-2.0)"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "license": [
            "https://spdx.org/licenses/Apache-2.0",
            "https://spdx.org/licenses/MIT",
        ],
    }

    package_json = rb"""{
  "license": "MIT OR (LGPL-2.1-only OR BSD-3-Clause)"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "license": [
            "https://spdx.org/licenses/BSD-3-Clause",
            "https://spdx.org/licenses/LGPL-2.1-only",
            "https://spdx.org/licenses/MIT",
        ],
    }

    package_json = rb"""{
  "license": "MIT OR LGPL-2.1-only OR BSD-3-Clause"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "license": [
            "https://spdx.org/licenses/BSD-3-Clause",
            "https://spdx.org/licenses/LGPL-2.1-only",
            "https://spdx.org/licenses/MIT",
        ],
    }


def test_unsupported_spdx_expressions():
    package_json = rb"""{
  "license": "MIT AND Apache-2.0"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    package_json = rb"""{
  "license": "(MIT AND Apache-2.0)"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    package_json = rb"""{
  "license": "MIT AND Apache-2.0 OR GPL-3.0-only"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    package_json = rb"""{
  "license": "MIT AND (LGPL-2.1-only OR BSD-3-Clause)"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    package_json = rb"""{
  "license": "MIT AND Apache-2.0 WITH Bison-exception-2.2"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    package_json = rb"""{
  "license": "MIT OR Apache-2.0 WITH Bison-exception-2.2"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    package_json = rb"""{
  "license": "MIT OR (LGPL-2.1-only WITH Bison-exception-2.2)"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    package_json = rb"""{
  "license": "MIT AND (LGPL-2.1-only WITH Bison-exception-2.2)"
}"""
    result = MAPPINGS["NpmMapping"]().translate(package_json)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.indexer.storage.model import BaseRow, ContentLicenseRow


def test_unique_key__no_tool_dict():
    with pytest.raises(ValueError, match="indexer_configuration_id"):
        BaseRow(id=12, indexer_configuration_id=34).unique_key()
    with pytest.raises(ValueError, match="indexer_configuration_id"):
        ContentLicenseRow(
            id=12, indexer_configuration_id=34, license="BSD"
        ).unique_key()


def test_unique_key():
    assert BaseRow(
        id=12, tool={"id": 34, "name": "foo", "version": "1.2.3", "configuration": {}}
    ).unique_key() == {
        "id": 12,
        "tool_name": "foo",
        "tool_version": "1.2.3",
        "tool_configuration": "{}",
    }

    assert ContentLicenseRow(
        id=12,
        tool={"id": 34, "name": "foo", "version": "1.2.3", "configuration": {}},
        license="BSD",
    ).unique_key() == {
        "id": 12,
        "license": "BSD",
        "tool_name": "foo",
        "tool_version": "1.2.3",
        "tool_configuration": "{}",
    }

    assert ContentLicenseRow(
        id=12,
        tool={
            "id": 34,
            "name": "foo",
            "version": "1.2.3",
            "configuration": {"foo": 1, "bar": 2},
        },
        license="BSD",
    ).unique_key() == {
        "id": 12,
        "license": "BSD",
        "tool_name": "foo",
        "tool_version": "1.2.3",
        "tool_configuration": '{"bar": 2, "foo": 1}',
    }



# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import csv
import itertools
import json
import os.path
import re
from typing import Any, Dict, List, Set, TextIO, Tuple

from pyld import jsonld
import rdflib

import swh.indexer
from swh.indexer.namespaces import ACTIVITYSTREAMS, CODEMETA, FORGEFED, SCHEMA

_DATA_DIR = os.path.join(os.path.dirname(swh.indexer.__file__), "data")

CROSSWALK_TABLE_PATH = os.path.join(_DATA_DIR, "codemeta", "crosswalk.csv")

CODEMETA_CONTEXT_PATH = os.path.join(_DATA_DIR, "codemeta", "codemeta.jsonld")


with open(CODEMETA_CONTEXT_PATH) as fd:
    CODEMETA_CONTEXT = json.load(fd)

_EMPTY_PROCESSED_CONTEXT: Any = {"mappings": {}}
_PROCESSED_CODEMETA_CONTEXT = jsonld.JsonLdProcessor().process_context(
    _EMPTY_PROCESSED_CONTEXT, CODEMETA_CONTEXT, None
)

CODEMETA_CONTEXT_URL = "https://doi.org/10.5063/schema/codemeta-2.0"
CODEMETA_ALTERNATE_CONTEXT_URLS = {
    ("https://raw.githubusercontent.com/codemeta/codemeta/master/codemeta.jsonld")
}


PROPERTY_BLACKLIST = {
    # CodeMeta properties that we cannot properly represent.
    SCHEMA.softwareRequirements,
    CODEMETA.softwareSuggestions,
    # Duplicate of 'author'
    SCHEMA.creator,
}

_codemeta_field_separator = re.compile(r"\s*[,/]\s*")


def make_absolute_uri(local_name):
    """Parses codemeta.jsonld, and returns the @id of terms it defines.

    >>> make_absolute_uri("name")
    'http://schema.org/name'
    >>> make_absolute_uri("downloadUrl")
    'http://schema.org/downloadUrl'
    >>> make_absolute_uri("referencePublication")
    'https://codemeta.github.io/terms/referencePublication'
    """
    uri = jsonld.JsonLdProcessor.get_context_value(
        _PROCESSED_CODEMETA_CONTEXT, local_name, "@id"
    )
    assert uri.startswith(("@", CODEMETA, SCHEMA)), (local_name, uri)
    return uri


def read_crosstable(fd: TextIO) -> Tuple[Set[str], Dict[str, Dict[str, rdflib.URIRef]]]:
    """
    Given a file-like object to a `CodeMeta crosswalk table` (either the main
    cross-table with all columns, or an auxiliary table with just the CodeMeta
    column and one ecosystem-specific table); returns a list of all CodeMeta
    terms, and a dictionary ``{ecosystem: {ecosystem_term: codemeta_term}}``

    .. _CodeMeta crosswalk table: <https://codemeta.github.io/crosswalk/
    """
    reader = csv.reader(fd)
    try:
        header = next(reader)
    except StopIteration:
        raise ValueError("empty file")

    data_sources = set(header) - {"Parent Type", "Property", "Type", "Description"}

    codemeta_translation: Dict[str, Dict[str, rdflib.URIRef]] = {
        data_source: {} for data_source in data_sources
    }
    terms = set()

    for line in reader:  # For each canonical name
        local_name = dict(zip(header, line))["Property"]
        if not local_name:
            continue
        canonical_name = make_absolute_uri(local_name)
        if rdflib.URIRef(canonical_name) in PROPERTY_BLACKLIST:
            continue
        terms.add(canonical_name)
        for col, value in zip(header, line):  # For each cell in the row
            if col in data_sources:
                # If that's not the parentType/property/type/description
                for local_name in _codemeta_field_separator.split(value):
                    # For each of the data source's properties that maps
                    # to this canonical name
                    if local_name.strip():
                        codemeta_translation[col][local_name.strip()] = rdflib.URIRef(
                            canonical_name
                        )

    return (terms, codemeta_translation)


with open(CROSSWALK_TABLE_PATH) as fd:
    (CODEMETA_TERMS, CROSSWALK_TABLE) = read_crosstable(fd)


def _document_loader(url, options=None):
    """Document loader for pyld.

    Reads the local codemeta.jsonld file instead of fetching it
    from the Internet every single time."""
    if (
        url.lower() == CODEMETA_CONTEXT_URL.lower()
        or url in CODEMETA_ALTERNATE_CONTEXT_URLS
    ):
        return {
            "contextUrl": None,
            "documentUrl": url,
            "document": CODEMETA_CONTEXT,
        }
    elif url == CODEMETA:
        raise Exception(
            "{} is CodeMeta's URI, use {} as context url".format(
                CODEMETA, CODEMETA_CONTEXT_URL
            )
        )
    else:
        raise Exception(url)


def compact(doc, forgefed: bool):
    """Same as `pyld.jsonld.compact`, but in the context of CodeMeta.

    Args:
        forgefed: Whether to add ForgeFed and ActivityStreams as compact URIs.
          This is typically used for extrinsic metadata documents, which frequently
          use properties from these namespaces.
    """
    contexts: List[Any] = [CODEMETA_CONTEXT_URL]
    if forgefed:
        contexts.append({"as": str(ACTIVITYSTREAMS), "forge": str(FORGEFED)})
    return jsonld.compact(doc, contexts, options={"documentLoader": _document_loader})


def expand(doc):
    """Same as `pyld.jsonld.expand`, but in the context of CodeMeta."""
    return jsonld.expand(doc, options={"documentLoader": _document_loader})


def merge_documents(documents):
    """Takes a list of metadata dicts, each generated from a different
    metadata file, and merges them.

    Removes duplicates, if any."""
    documents = list(itertools.chain.from_iterable(map(expand, documents)))
    merged_document = collections.defaultdict(list)
    for document in documents:
        for key, values in document.items():
            if key == "@id":
                # @id does not get expanded to a list
                value = values

                # Only one @id is allowed, move it to sameAs
                if "@id" not in merged_document:
                    merged_document["@id"] = value
                elif value != merged_document["@id"]:
                    if value not in merged_document[SCHEMA.sameAs]:
                        merged_document[SCHEMA.sameAs].append(value)
            else:
                for value in values:
                    if isinstance(value, dict) and set(value) == {"@list"}:
                        # Value is of the form {'@list': [item1, item2]}
                        # instead of the usual [item1, item2].
                        # We need to merge the inner lists (and mostly
                        # preserve order).
                        merged_value = merged_document.setdefault(key, {"@list": []})
                        for subvalue in value["@list"]:
                            # merged_value must be of the form
                            # {'@list': [item1, item2]}; as it is the same
                            # type as value, which is an @list.
                            if subvalue not in merged_value["@list"]:
                                merged_value["@list"].append(subvalue)
                    elif value not in merged_document[key]:
                        merged_document[key].append(value)

    # XXX: we should set forgefed=True when merging extrinsic_metadata documents.
    # however, this function is only used to merge multiple files of the same
    # directory (which is only for intrinsic-metadata), so it is not an issue for now
    return compact(merged_document, forgefed=False)



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.indexer.metadata_detector import detect_metadata
from swh.indexer.metadata_dictionary import MAPPINGS


def test_compute_metadata_nuget():
    raw_content = b"""<?xml version="1.0" encoding="utf-8"?>
    <package xmlns="http://schemas.microsoft.com/packaging/2010/07/nuspec.xsd">
        <metadata>
            <id>sample</id>
            <version>1.2.3</version>
            <authors>Kim Abercrombie, Franck Halmaert</authors>
            <description>Sample exists only to show a sample .nuspec file.</description>
            <summary>Summary is being deprecated. Use description instead.</summary>
            <projectUrl>http://example.org/</projectUrl>
            <repository type="git" url="https://github.com/NuGet/NuGet.Client.git"/>
            <license type="expression">MIT</license>
            <licenseUrl>https://raw.github.com/timrwood/moment/master/LICENSE</licenseUrl>
            <dependencies>
                <dependency id="another-package" version="3.0.0" />
                <dependency id="yet-another-package" version="1.0.0" />
            </dependencies>
            <releaseNotes>
                See the [changelog](https://github.com/httpie/httpie/releases/tag/3.2.0).
            </releaseNotes>
            <tags>python3 java cpp search-tag</tags>
        </metadata>
        <files>
            <file src="bin\\Debug\\*.dll" target="lib" />
        </files>
    </package>"""

    result = MAPPINGS["NuGetMapping"]().translate(raw_content)

    assert set(result.pop("keywords")) == {
        "python3",
        "java",
        "cpp",
        "search-tag",
    }, result

    assert set(result.pop("license")) == {
        "https://spdx.org/licenses/MIT",
        "https://raw.github.com/timrwood/moment/master/LICENSE",
    }, result

    assert set(result.pop("description")) == {
        "Sample exists only to show a sample .nuspec file.",
        "Summary is being deprecated. Use description instead.",
    }, result

    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [
            {"type": "Person", "name": "Kim Abercrombie"},
            {"type": "Person", "name": "Franck Halmaert"},
        ],
        "codeRepository": "https://github.com/NuGet/NuGet.Client.git",
        "url": "http://example.org/",
        "version": "1.2.3",
        "schema:releaseNotes": (
            "See the [changelog](https://github.com/httpie/httpie/releases/tag/3.2.0)."
        ),
    }

    assert result == expected


@pytest.mark.parametrize(
    "filename",
    [b"package_name.nuspec", b"number_5.nuspec", b"CAPS.nuspec", b"\x8anan.nuspec"],
)
def test_detect_metadata_package_nuspec(filename):
    df = [
        {
            "sha1_git": b"abc",
            "name": b"example.json",
            "target": b"abc",
            "length": 897,
            "status": "visible",
            "type": "file",
            "perms": 33188,
            "dir_id": b"dir_a",
            "sha1": b"bcd",
        },
        {
            "sha1_git": b"aab",
            "name": filename,
            "target": b"aab",
            "length": 712,
            "status": "visible",
            "type": "file",
            "perms": 33188,
            "dir_id": b"dir_a",
            "sha1": b"cde",
        },
    ]
    results = detect_metadata(df)

    expected_results = {"NuGetMapping": [b"cde"]}
    assert expected_results == results


def test_normalize_license_multiple_licenses_or_delimiter():
    raw_content = raw_content = b"""<?xml version="1.0" encoding="utf-8"?>
    <package xmlns="http://schemas.microsoft.com/packaging/2010/07/nuspec.xsd">
        <metadata>
            <license type="expression">BitTorrent-1.0 or GPL-3.0-with-GCC-exception</license>
        </metadata>
        <files>
            <file src="bin\\Debug\\*.dll" target="lib" />
        </files>
    </package>"""
    result = MAPPINGS["NuGetMapping"]().translate(raw_content)
    assert set(result.pop("license")) == {
        "https://spdx.org/licenses/BitTorrent-1.0",
        "https://spdx.org/licenses/GPL-3.0-with-GCC-exception",
    }
    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    assert result == expected


def test_normalize_license_unsupported_delimiter():
    raw_content = raw_content = b"""<?xml version="1.0" encoding="utf-8"?>
    <package xmlns="http://schemas.microsoft.com/packaging/2010/07/nuspec.xsd">
        <metadata>
            <license type="expression">(MIT)</license>
        </metadata>
        <files>
            <file src="bin\\Debug\\*.dll" target="lib" />
        </files>
    </package>"""
    result = MAPPINGS["NuGetMapping"]().translate(raw_content)
    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }

    assert result == expected


def test_copyrightNotice_absolute_uri_property():
    raw_content = raw_content = b"""<?xml version="1.0" encoding="utf-8"?>
    <package xmlns="http://schemas.microsoft.com/packaging/2010/07/nuspec.xsd">
        <metadata>
            <copyright>Copyright 2017-2022</copyright>
            <language>en-us</language>
        </metadata>
        <files>
            <file src="bin\\Debug\\*.dll" target="lib" />
        </files>
    </package>"""
    result = MAPPINGS["NuGetMapping"]().translate(raw_content)
    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "schema:copyrightNotice": "Copyright 2017-2022",
        "schema:inLanguage": "en-us",
    }

    assert result == expected



# Copyright (C) 2017-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict
import unittest
from unittest.mock import patch

import pytest

from swh.indexer import fossology_license
from swh.indexer.fossology_license import FossologyLicenseIndexer, compute_license
from swh.indexer.storage.model import ContentLicenseRow
from swh.indexer.tests.utils import (
    BASE_TEST_CONFIG,
    RAW_CONTENT_IDS,
    SHA1_TO_LICENSES,
    CommonContentIndexerTest,
    fill_obj_storage,
    fill_storage,
    filter_dict,
    mock_compute_license,
)


class BasicTest(unittest.TestCase):
    @patch("swh.indexer.fossology_license.subprocess")
    def test_compute_license(self, mock_subprocess):
        """Computing licenses from a raw content should return results"""
        for path, intermediary_result, output in [
            (b"some/path", None, []),
            (b"some/path/2", [], []),
            (b"other/path", " contains license(s) GPL,AGPL", ["GPL", "AGPL"]),
        ]:
            mock_subprocess.check_output.return_value = intermediary_result

            actual_result = compute_license(path)

            self.assertEqual(
                actual_result,
                {
                    "licenses": output,
                    "path": path,
                },
            )


CONFIG: Dict[str, Any] = {
    **BASE_TEST_CONFIG,
    "workdir": "/tmp",
    "tools": {
        "name": "nomos",
        "version": "3.1.0rc2-31-ga2cbb8c",
        "configuration": {
            "command_line": "nomossa <filepath>",
        },
    },
}

RANGE_CONFIG = dict(list(CONFIG.items()) + [("write_batch_size", 100)])


class TestFossologyLicenseIndexer(CommonContentIndexerTest, unittest.TestCase):
    """Fossology license indexer test scenarios:

    - Known sha1s in the input list have their data indexed
    - Unknown sha1 in the input list are not indexed

    """

    def get_indexer_results(self, ids):
        yield from self.idx_storage.content_fossology_license_get(ids)

    def setUp(self):
        super().setUp()
        # replace actual license computation with a mock
        self.orig_compute_license = fossology_license.compute_license
        fossology_license.compute_license = mock_compute_license

        self.indexer = FossologyLicenseIndexer(CONFIG)
        self.indexer.catch_exceptions = False
        self.idx_storage = self.indexer.idx_storage
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

        self.id0, self.id1, self.id2 = RAW_CONTENT_IDS

        tool = {k.replace("tool_", ""): v for (k, v) in self.indexer.tool.items()}

        # then
        self.expected_results = [
            *[
                ContentLicenseRow(id=self.id0, tool=tool, license=license)
                for license in SHA1_TO_LICENSES[self.id0]
            ],
            *[
                ContentLicenseRow(id=self.id1, tool=tool, license=license)
                for license in SHA1_TO_LICENSES[self.id1]
            ],
            *[],  # self.id2
        ]

    def tearDown(self):
        super().tearDown()
        fossology_license.compute_license = self.orig_compute_license


def test_fossology_w_no_tool():
    with pytest.raises(ValueError):
        FossologyLicenseIndexer(config=filter_dict(CONFIG, "tools"))



# Copyright (C) 2020-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, Iterable, Optional

try:
    from swh.journal.writer import JournalWriterInterface, get_journal_writer
except ImportError:
    get_journal_writer = None  # type: ignore
    # mypy limitation, see https://github.com/python/mypy/issues/1153

from .model import BaseRow


class JournalWriter:
    """Journal writer storage collaborator. It's in charge of adding objects to
    the journal.

    """

    journal: Optional[JournalWriterInterface]

    def __init__(self, journal_writer: Dict[str, Any]):
        """
        Args:
            journal_writer: configuration passed to
                            `swh.journal.writer.get_journal_writer`
        """
        if journal_writer:
            if get_journal_writer is None:
                raise EnvironmentError(
                    "You need the swh.journal package to use the "
                    "journal_writer feature"
                )
            self.journal = get_journal_writer(
                **journal_writer,
                value_sanitizer=lambda object_type, value_dict: value_dict,
            )
        else:
            self.journal = None

    def write_additions(self, obj_type, entries: Iterable[BaseRow]) -> None:
        if not self.journal:
            return

        translated = []

        for entry in entries:
            assert entry.object_type == obj_type  # type: ignore

            # ids are internal to the database and should not be sent to postgresql
            if entry.indexer_configuration_id is not None:
                raise ValueError(
                    f"{entry} passed to JournalWriter.write_additions has "
                    f"indexer_configuration_id instead of full tool dict"
                )
            assert entry.tool, "Missing both indexer_configuration_id and tool dict"
            if "id" in entry.tool:
                raise ValueError(
                    f"{entry} passed to JournalWriter.write_additions "
                    f"contains a tool id"
                )

            translated.append(entry)

        # write to kafka
        self.journal.write_additions(obj_type, translated)



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import textwrap

import pytest

from swh.indexer.bibtex import cff_to_bibtex, codemeta_to_bibtex
from swh.model.swhids import QualifiedSWHID


def test_empty():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        }
    ) == textwrap.dedent(
        """\
            @software{REPLACEME
            }
            """
    )


def test_minimal():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {"name": "Jane Doe"},
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
        }
    ) == textwrap.dedent(
        """\
        @software{REPLACEME,
            author = "Doe, Jane",
            date = "2023-10-10",
            year = "2023",
            month = "10",
            title = "Example Software",
            url = "http://example.org/"
        }
        """
    )


def test_empty_author_list():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": [],
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
        }
    ) == textwrap.dedent(
        """\
        @software{REPLACEME,
            date = "2023-10-10",
            year = "2023",
            month = "10",
            title = "Example Software",
            url = "http://example.org/"
        }
        """
    )


@pytest.mark.parametrize("key", ["version", "softwareVersion"])
def test_version_minimal(key):
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {"name": "Jane Doe"},
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
            key: "1.2.3",
        }
    ) == textwrap.dedent(
        """\
        @softwareversion{REPLACEME,
            author = "Doe, Jane",
            date = "2023-10-10",
            year = "2023",
            month = "10",
            title = "Example Software",
            url = "http://example.org/",
            version = "1.2.3"
        }
        """
    )


def test_id():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {"name": "Jane Doe"},
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
            "identifier": "example-software",
        }
    ) == textwrap.dedent(
        """\
        @software{example-software,
            author = "Doe, Jane",
            date = "2023-10-10",
            year = "2023",
            month = "10",
            title = "Example Software",
            url = "http://example.org/"
        }
        """
    )


@pytest.mark.parametrize("key", ["@id", "id", "identifier"])
def test_id_url(key):
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {"name": "Jane Doe"},
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
            key: "http://example.org/example-software",
        }
    ) == textwrap.dedent(
        """\
        @software{REPLACEME,
            author = "Doe, Jane",
            date = "2023-10-10",
            year = "2023",
            month = "10",
            title = "Example Software",
            url = "http://example.org/"
        }
        """
    )


@pytest.mark.parametrize("key", ["@id", "id", "identifier"])
def test_id_doi(key):
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {"name": "Jane Doe"},
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
            key: "https://doi.org/10.1000/182",
        }
    ) == textwrap.dedent(
        """\
        @software{REPLACEME,
            author = "Doe, Jane",
            date = "2023-10-10",
            year = "2023",
            month = "10",
            doi = "https://doi.org/10.1000/182",
            title = "Example Software",
            url = "http://example.org/"
        }
        """
    )


def test_license():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {"name": "Jane Doe"},
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
            "license": "https://spdx.org/licenses/Apache-2.0",
        }
    ) == textwrap.dedent(
        """\
        @software{REPLACEME,
            author = "Doe, Jane",
            license = "Apache-2.0",
            date = "2023-10-10",
            year = "2023",
            month = "10",
            title = "Example Software",
            url = "http://example.org/"
        }
        """
    )


def test_licenses():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {"name": "Jane Doe"},
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
            "license": [
                "https://spdx.org/licenses/Apache-2.0",
                "https://spdx.org/licenses/GPL-3.0",
            ],
        }
    ) == textwrap.dedent(
        """\
        @software{REPLACEME,
            author = "Doe, Jane",
            license = "Apache-2.0 and GPL-3.0",
            date = "2023-10-10",
            year = "2023",
            month = "10",
            title = "Example Software",
            url = "http://example.org/"
        }
        """
    )


def test_organization_author():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {
                "@type": "Organization",
                "name": "Example University",
            },
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
        }
    ) == textwrap.dedent(
        """\
        @software{REPLACEME,
            author = "Example University",
            date = "2023-10-10",
            year = "2023",
            month = "10",
            title = "Example Software",
            url = "http://example.org/"
        }
        """
    )


def test_affiliation():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "author": {
                "name": "Jane Doe",
                "affiliation": {"@type": "Organization", "name": "Example University"},
            },
            "name": "Example Software",
            "url": "http://example.org/",
            "datePublished": "2023-10-10",
        }
    ) == textwrap.dedent(
        """\
        @software{REPLACEME,
            author = "Doe, Jane",
            organization = "Example University",
            date = "2023-10-10",
            year = "2023",
            month = "10",
            title = "Example Software",
            url = "http://example.org/"
        }
        """
    )


def test_cff_empty():
    assert cff_to_bibtex("") == textwrap.dedent(
        """\
        @software{REPLACEME
        }
        """
    )


def test_cff_invalid():
    assert cff_to_bibtex("foo") == textwrap.dedent(
        """\
        @software{REPLACEME
        }
        """
    )


def test_cff_minimal():
    assert (
        cff_to_bibtex(
            """
cff-version: 1.2.0
message: "If you use this software, please cite it as below."
authors:
  - family-names: Druskat
    given-names: Stephan
title: "My Research Software"
date-released: 2021-08-11
url: "http://example.org/"
            """
        )
        == textwrap.dedent(
            """\
            @software{REPLACEME,
                author = "Druskat, Stephan",
                date = "2021-08-11",
                year = "2021",
                month = "08",
                title = "My Research Software",
                url = "http://example.org/"
            }
            """
        )
    )


def test_swhid_type_snp():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        },
        QualifiedSWHID.from_string(
            "swh:1:snp:da39a3ee5e6b4b0d3255bfef95601890afd80709"
        ),
    ) == textwrap.dedent(
        """\
        @software{REPLACEME,
            swhid = "swh:1:snp:da39a3ee5e6b4b0d3255bfef95601890afd80709"
        }
        """
    )


def test_swhid_type_rev():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        },
        QualifiedSWHID.from_string(
            "swh:1:rev:5b909292bcfe6099d726c0b5194165c72f93b767"
        ),
    ) == textwrap.dedent(
        """\
        @softwareversion{REPLACEME,
            swhid = "swh:1:rev:5b909292bcfe6099d726c0b5194165c72f93b767"
        }
        """
    )


def test_swhid_type_cnt():
    assert codemeta_to_bibtex(
        {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        },
        QualifiedSWHID.from_string(
            "swh:1:cnt:5b909292bcfe6099d726c0b5194165c72f93b767;lines=5-10"
        ),
    ) == textwrap.dedent(
        """\
        @codefragment{REPLACEME,
            swhid = "swh:1:cnt:5b909292bcfe6099d726c0b5194165c72f93b767;lines=5-10"
        }
        """
    )



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from os.path import join

import pytest

from swh.indexer.storage import get_indexer_storage
from swh.indexer.storage.model import ContentLicenseRow, ContentMimetypeRow
from swh.indexer.tests.conftest import idx_storage_postgresql
from swh.model.hashutil import hash_to_bytes

from . import SQL_DIR
from .generate_data_test import FOSSOLOGY_LICENSES, MIMETYPE_OBJECTS, TOOLS

DUMP_FILES = join(SQL_DIR, "*.sql")


class DataObj(dict):
    def __getattr__(self, key):
        return self.__getitem__(key)

    def __setattr__(self, key, value):
        return self.__setitem__(key, value)


@pytest.fixture
def swh_indexer_storage_with_data(swh_indexer_storage):
    data = DataObj()
    tools = {
        tool["tool_name"]: {
            "id": tool["id"],
            "name": tool["tool_name"],
            "version": tool["tool_version"],
            "configuration": tool["tool_configuration"],
        }
        for tool in swh_indexer_storage.indexer_configuration_add(TOOLS)
    }
    data.tools = tools
    data.sha1_1 = hash_to_bytes("34973274ccef6ab4dfaaf86599792fa9c3fe4689")
    data.sha1_2 = hash_to_bytes("61c2b3a30496d329e21af70dd2d7e097046d07b7")
    data.directory_id_1 = hash_to_bytes("7026b7c1a2af56521e951c01ed20f255fa054238")
    data.directory_id_2 = hash_to_bytes("7026b7c1a2af56521e9587659012345678904321")
    data.directory_id_3 = hash_to_bytes("7026b7c1a2af56521e9587659012345678904320")
    data.origin_url_1 = "file:///dev/0/zero"  # 44434341
    data.origin_url_2 = "file:///dev/1/one"  # 44434342
    data.origin_url_3 = "file:///dev/2/two"  # 54974445
    data.mimetypes = [
        ContentMimetypeRow(indexer_configuration_id=tools["file"]["id"], **mimetype_obj)
        for mimetype_obj in MIMETYPE_OBJECTS
    ]
    swh_indexer_storage.content_mimetype_add(data.mimetypes)
    data.fossology_licenses = [
        ContentLicenseRow(
            id=fossology_obj["id"],
            indexer_configuration_id=tools["nomos"]["id"],
            license=license,
        )
        for fossology_obj in FOSSOLOGY_LICENSES
        for license in fossology_obj["licenses"]
    ]
    swh_indexer_storage._test_data = data

    return (swh_indexer_storage, data)


swh_indexer_storage_postgresql = idx_storage_postgresql


@pytest.fixture
def swh_indexer_storage(swh_indexer_storage_postgresql):
    return get_indexer_storage(
        "local",
        db=swh_indexer_storage_postgresql.info.dsn,
        journal_writer={
            "cls": "memory",
        },
    )



# Copyright (C) 2015-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import psycopg2
import pytest

from swh.core.api import RemoteException, TransientRemoteException
from swh.indexer.storage import get_indexer_storage
from swh.indexer.storage.api.client import RemoteStorage
import swh.indexer.storage.api.server as server

from .test_storage import *  # noqa


@pytest.fixture
def app_server(swh_indexer_storage_postgresql):
    server.storage = get_indexer_storage(
        "local",
        db=swh_indexer_storage_postgresql.info.dsn,
        journal_writer={
            "cls": "memory",
        },
    )
    yield server


@pytest.fixture
def app(app_server):
    return app_server.app


@pytest.fixture
def swh_rpc_client_class():
    # these are needed for the swh_indexer_storage_with_data fixture
    assert hasattr(RemoteStorage, "indexer_configuration_add")
    assert hasattr(RemoteStorage, "content_mimetype_add")
    return RemoteStorage


@pytest.fixture
def swh_indexer_storage(swh_rpc_client, app_server):
    # This version of the swh_storage fixture uses the swh_rpc_client fixture
    # to instantiate a RemoteStorage (see swh_rpc_client_class above) that
    # proxies, via the swh.core RPC mechanism, the local (in memory) storage
    # configured in the app fixture above.
    #
    # Also note that, for the sake of
    # making it easier to write tests, the in-memory journal writer of the
    # in-memory backend storage is attached to the RemoteStorage as its
    # journal_writer attribute.
    storage = swh_rpc_client

    journal_writer = getattr(storage, "journal_writer", None)
    storage.journal_writer = app_server.storage.journal_writer
    yield storage
    storage.journal_writer = journal_writer


def test_exception(app_server, swh_indexer_storage, mocker):
    """Checks the client re-raises unknown exceptions as a :exc:`RemoteException`"""
    assert swh_indexer_storage.content_mimetype_get([b"\x01" * 20]) == []
    mocker.patch.object(
        app_server.storage,
        "content_mimetype_get",
        side_effect=ValueError("crash"),
    )
    with pytest.raises(RemoteException) as e:
        swh_indexer_storage.content_mimetype_get([b"\x01" * 20])
    assert not isinstance(e, TransientRemoteException)


def test_operationalerror_exception(app_server, swh_indexer_storage, mocker):
    """Checks the client re-raises as a :exc:`TransientRemoteException`
    rather than the base :exc:`RemoteException`; so the retrying proxy
    retries for longer."""
    assert swh_indexer_storage.content_mimetype_get([b"\x01" * 20]) == []
    mocker.patch.object(
        app_server.storage,
        "content_mimetype_get",
        side_effect=psycopg2.errors.AdminShutdown("cluster is shutting down"),
    )
    with pytest.raises(RemoteException) as excinfo:
        swh_indexer_storage.content_mimetype_get([b"\x01" * 20])
    assert isinstance(excinfo.value, TransientRemoteException)


def test_querycancelled_exception(app_server, swh_indexer_storage, mocker):
    """Checks the client re-raises as a :exc:`TransientRemoteException`
    rather than the base :exc:`RemoteException`; so the retrying proxy
    retries for longer."""
    assert swh_indexer_storage.content_mimetype_get([b"\x01" * 20]) == []
    mocker.patch.object(
        app_server.storage,
        "content_mimetype_get",
        side_effect=psycopg2.errors.QueryCanceled("too big!"),
    )
    with pytest.raises(RemoteException) as excinfo:
        swh_indexer_storage.content_mimetype_get([b"\x01" * 20])
    assert not isinstance(excinfo.value, TransientRemoteException)



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from functools import reduce
import re
from typing import Any, Dict, List

import attr
from click.testing import CliRunner
from confluent_kafka import Consumer
import pytest

from swh.indexer import fossology_license
from swh.indexer.cli import indexer_cli_group
from swh.indexer.storage.interface import IndexerStorageInterface
from swh.indexer.storage.model import (
    ContentLicenseRow,
    ContentMimetypeRow,
    DirectoryIntrinsicMetadataRow,
    OriginExtrinsicMetadataRow,
    OriginIntrinsicMetadataRow,
)
from swh.journal.writer import get_journal_writer
from swh.model.hashutil import hash_to_bytes
from swh.model.model import Content, Origin, OriginVisitStatus

from .test_metadata import GITHUB_REMD
from .utils import (
    DIRECTORY2,
    RAW_CONTENT_IDS,
    RAW_CONTENTS,
    REVISION,
    SHA1_TO_LICENSES,
    mock_compute_license,
)


def fill_idx_storage(idx_storage: IndexerStorageInterface, nb_rows: int) -> List[int]:
    tools: List[Dict[str, Any]] = [
        {
            "tool_name": "tool %d" % i,
            "tool_version": "0.0.1",
            "tool_configuration": {},
        }
        for i in range(2)
    ]
    tools = idx_storage.indexer_configuration_add(tools)

    origin_metadata = [
        OriginIntrinsicMetadataRow(
            id="file://dev/%04d" % origin_id,
            from_directory=hash_to_bytes("abcd{:0>36}".format(origin_id)),
            indexer_configuration_id=tools[origin_id % 2]["id"],
            metadata={"name": "origin %d" % origin_id},
            mappings=["mapping%d" % (origin_id % 10)],
        )
        for origin_id in range(nb_rows)
    ]
    directory_metadata = [
        DirectoryIntrinsicMetadataRow(
            id=hash_to_bytes("abcd{:0>36}".format(origin_id)),
            indexer_configuration_id=tools[origin_id % 2]["id"],
            metadata={"name": "origin %d" % origin_id},
            mappings=["mapping%d" % (origin_id % 10)],
        )
        for origin_id in range(nb_rows)
    ]

    idx_storage.directory_intrinsic_metadata_add(directory_metadata)
    idx_storage.origin_intrinsic_metadata_add(origin_metadata)

    return [tool["id"] for tool in tools]


def _origins_in_task_args(tasks):
    """Returns the set of origins contained in the arguments of the
    provided tasks (assumed to be of type index-origin-metadata)."""
    return reduce(
        set.union, (set(task["arguments"]["args"][0]) for task in tasks), set()
    )


def _assert_tasks_for_origins(tasks, origins):
    expected_kwargs = {}
    assert {task["type"] for task in tasks} == {"index-origin-metadata"}
    assert all(len(task["arguments"]["args"]) == 1 for task in tasks)
    for task in tasks:
        assert task["arguments"]["kwargs"] == expected_kwargs, task
    assert _origins_in_task_args(tasks) == set(["file://dev/%04d" % i for i in origins])


@pytest.fixture
def cli_runner():
    return CliRunner()


def test_cli_mapping_list(cli_runner, swh_config):
    result = cli_runner.invoke(
        indexer_cli_group,
        ["-C", swh_config, "mapping", "list"],
        catch_exceptions=False,
    )
    expected_output = "\n".join(
        [
            "cff",
            "codemeta",
            "composer",
            "gemspec",
            "gitea",
            "github",
            "json-sword-codemeta",
            "maven",
            "npm",
            "nuget",
            "pkg-info",
            "pubspec",
            "sword-codemeta",
            "",
        ]  # must be sorted for test to pass
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output


def test_cli_mapping_list_terms(cli_runner, swh_config):
    result = cli_runner.invoke(
        indexer_cli_group,
        ["-C", swh_config, "mapping", "list-terms"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert re.search(r"http://schema.org/url:\n.*npm", result.output)
    assert re.search(r"http://schema.org/url:\n.*codemeta", result.output)
    assert re.search(
        r"https://codemeta.github.io/terms/developmentStatus:\n\tcodemeta",
        result.output,
    )


def test_cli_mapping_list_terms_exclude(cli_runner, swh_config):
    result = cli_runner.invoke(
        indexer_cli_group,
        [
            "-C",
            swh_config,
            "mapping",
            "list-terms",
            "--exclude-mapping",
            "codemeta",
            "--exclude-mapping",
            "json-sword-codemeta",
            "--exclude-mapping",
            "sword-codemeta",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert re.search(r"http://schema.org/url:\n.*npm", result.output)
    assert not re.search(r"http://schema.org/url:\n.*codemeta", result.output)
    assert not re.search(
        r"https://codemeta.github.io/terms/developmentStatus:\n\tcodemeta",
        result.output,
    )


def now():
    return datetime.datetime.now(tz=datetime.timezone.utc)


def test_cli_journal_client_without_brokers(
    cli_runner, swh_config, kafka_prefix: str, kafka_server, consumer: Consumer
):
    """Without brokers configuration, the cli fails."""

    with pytest.raises(ValueError, match="brokers"):
        cli_runner.invoke(
            indexer_cli_group,
            [
                "-C",
                swh_config,
                "journal-client",
                "*",
            ],
            catch_exceptions=False,
        )


@pytest.mark.parametrize("indexer_name", ["origin_intrinsic_metadata", "*"])
def test_cli_journal_client_index__origin_intrinsic_metadata(
    cli_runner,
    swh_config,
    kafka_prefix: str,
    kafka_server,
    consumer: Consumer,
    idx_storage,
    storage,
    mocker,
    swh_indexer_config,
    indexer_name: str,
):
    """Test the 'swh indexer journal-client' cli tool."""
    journal_writer = get_journal_writer(
        "kafka",
        brokers=[kafka_server],
        prefix=kafka_prefix,
        client_id="test producer",
        value_sanitizer=lambda object_type, value: value,
        flush_timeout=3,  # fail early if something is going wrong
    )

    visit_statuses = [
        OriginVisitStatus(
            origin="file:///dev/zero",
            type="git",
            visit=1,
            date=now(),
            status="full",
            snapshot=None,
        ),
        OriginVisitStatus(
            origin="file:///dev/foobar",
            type="git",
            visit=2,
            date=now(),
            status="full",
            snapshot=None,
        ),
        OriginVisitStatus(
            origin="file:///tmp/spamegg",
            type="git",
            visit=3,
            date=now(),
            status="full",
            snapshot=None,
        ),
        OriginVisitStatus(
            origin="file:///dev/0002",
            type="git",
            visit=6,
            date=now(),
            status="full",
            snapshot=None,
        ),
        OriginVisitStatus(  # will be filtered out due to its 'partial' status
            origin="file:///dev/0000",
            type="git",
            visit=4,
            date=now(),
            status="partial",
            snapshot=None,
        ),
        OriginVisitStatus(  # will be filtered out due to its 'ongoing' status
            origin="file:///dev/0001",
            type="git",
            visit=5,
            date=now(),
            status="ongoing",
            snapshot=None,
        ),
    ]

    journal_writer.write_additions("origin_visit_status", visit_statuses)
    visit_statuses_full = [vs for vs in visit_statuses if vs.status == "full"]
    storage.revision_add([REVISION])

    mocker.patch(
        "swh.indexer.metadata.get_head_swhid",
        return_value=REVISION.swhid(),
    )

    mocker.patch(
        "swh.indexer.metadata.DirectoryMetadataIndexer.index",
        return_value=[
            DirectoryIntrinsicMetadataRow(
                id=DIRECTORY2.id,
                indexer_configuration_id=1,
                mappings=["cff"],
                metadata={"foo": "bar"},
            )
        ],
    )
    result = cli_runner.invoke(
        indexer_cli_group,
        [
            "-C",
            swh_config,
            "journal-client",
            indexer_name,
            "--broker",
            kafka_server,
            "--prefix",
            kafka_prefix,
            "--group-id",
            "test-consumer",
            "--stop-after-objects",
            len(visit_statuses),
        ],
        catch_exceptions=False,
    )

    # Check the output
    expected_output = "Done.\n"
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    results = idx_storage.origin_intrinsic_metadata_get(
        [status.origin for status in visit_statuses]
    )
    expected_results = [
        OriginIntrinsicMetadataRow(
            id=status.origin,
            from_directory=DIRECTORY2.id,
            tool={"id": 1, **swh_indexer_config["tools"]},
            mappings=["cff"],
            metadata={"foo": "bar"},
        )
        for status in sorted(visit_statuses_full, key=lambda r: r.origin)
    ]
    assert sorted(results, key=lambda r: r.id) == expected_results


@pytest.mark.parametrize("indexer_name", ["extrinsic_metadata", "*"])
def test_cli_journal_client_index__origin_extrinsic_metadata(
    cli_runner,
    swh_config,
    kafka_prefix: str,
    kafka_server,
    consumer: Consumer,
    idx_storage,
    storage,
    mocker,
    swh_indexer_config,
    indexer_name: str,
):
    """Test the 'swh indexer journal-client' cli tool."""
    journal_writer = get_journal_writer(
        "kafka",
        brokers=[kafka_server],
        prefix=kafka_prefix,
        client_id="test producer",
        value_sanitizer=lambda object_type, value: value,
        flush_timeout=3,  # fail early if something is going wrong
    )

    origin = Origin("http://example.org/repo.git")
    storage.origin_add([origin])
    raw_extrinsic_metadata = attr.evolve(GITHUB_REMD, target=origin.swhid())
    raw_extrinsic_metadata = attr.evolve(
        raw_extrinsic_metadata, id=raw_extrinsic_metadata.compute_hash()
    )
    journal_writer.write_additions("raw_extrinsic_metadata", [raw_extrinsic_metadata])

    result = cli_runner.invoke(
        indexer_cli_group,
        [
            "-C",
            swh_config,
            "journal-client",
            indexer_name,
            "--broker",
            kafka_server,
            "--prefix",
            kafka_prefix,
            "--group-id",
            "test-consumer",
            "--stop-after-objects",
            1,
        ],
        catch_exceptions=False,
    )

    # Check the output
    expected_output = "Done.\n"
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    results = idx_storage.origin_extrinsic_metadata_get([origin.url])
    expected_results = [
        OriginExtrinsicMetadataRow(
            id=origin.url,
            from_remd_id=raw_extrinsic_metadata.id,
            tool={"id": 1, **swh_indexer_config["tools"]},
            mappings=["github"],
            metadata={
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "id": "http://example.org/",
                "type": "https://forgefed.org/ns#Repository",
                "name": "test software",
            },
        )
    ]
    assert sorted(results, key=lambda r: r.id) == expected_results


def test_cli_journal_client_index__content_mimetype(
    cli_runner,
    swh_config,
    kafka_prefix: str,
    kafka_server,
    consumer: Consumer,
    idx_storage,
    obj_storage,
    storage,
    mocker,
    swh_indexer_config,
):
    """Test the 'swh indexer journal-client' cli tool."""
    journal_writer = get_journal_writer(
        "kafka",
        brokers=[kafka_server],
        prefix=kafka_prefix,
        client_id="test producer",
        value_sanitizer=lambda object_type, value: value,
        flush_timeout=3,  # fail early if something is going wrong
    )

    contents = []
    expected_results = []
    content_ids = []
    for content_id, (raw_content, mimetypes, encoding) in RAW_CONTENTS.items():
        content = Content.from_data(raw_content)
        assert content_id == content.sha1

        contents.append(content)
        content_ids.append(content_id)

        # Older libmagic versions (e.g. buster: 1:5.35-4+deb10u2, bullseye: 1:5.39-3)
        # returns different results. This allows to deal with such a case when executing
        # tests on different environments machines (e.g. ci tox, ci debian, dev machine,
        # ...)
        all_mimetypes = mimetypes if isinstance(mimetypes, tuple) else [mimetypes]

        expected_results.extend(
            [
                ContentMimetypeRow(
                    id=content.sha1,
                    tool={"id": 1, **swh_indexer_config["tools"]},
                    mimetype=mimetype,
                    encoding=encoding,
                )
                for mimetype in all_mimetypes
            ]
        )

    assert len(contents) == len(RAW_CONTENTS)

    journal_writer.write_additions("content", contents)

    result = cli_runner.invoke(
        indexer_cli_group,
        [
            "-C",
            swh_config,
            "journal-client",
            "content_mimetype",
            "--broker",
            kafka_server,
            "--prefix",
            kafka_prefix,
            "--group-id",
            "test-consumer",
            "--stop-after-objects",
            len(contents),
        ],
        catch_exceptions=False,
    )

    # Check the output
    expected_output = "Done.\n"
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    results = idx_storage.content_mimetype_get(content_ids)
    assert len(results) == len(contents)
    for result in results:
        assert result in expected_results


def test_cli_journal_client_index__fossology_license(
    cli_runner,
    swh_config,
    kafka_prefix: str,
    kafka_server,
    consumer: Consumer,
    idx_storage,
    obj_storage,
    storage,
    mocker,
    swh_indexer_config,
):
    """Test the 'swh indexer journal-client' cli tool."""

    # Patch
    fossology_license.compute_license = mock_compute_license

    journal_writer = get_journal_writer(
        "kafka",
        brokers=[kafka_server],
        prefix=kafka_prefix,
        client_id="test producer",
        value_sanitizer=lambda object_type, value: value,
        flush_timeout=3,  # fail early if something is going wrong
    )

    tool = {"id": 1, **swh_indexer_config["tools"]}

    id0, id1, id2 = RAW_CONTENT_IDS

    contents = []
    content_ids = []
    expected_results = []
    for content_id, (raw_content, _, _) in RAW_CONTENTS.items():
        content = Content.from_data(raw_content)
        assert content_id == content.sha1

        contents.append(content)
        content_ids.append(content_id)

        expected_results.extend(
            [
                ContentLicenseRow(id=content_id, tool=tool, license=license)
                for license in SHA1_TO_LICENSES[content_id]
            ]
        )

    assert len(contents) == len(RAW_CONTENTS)

    journal_writer.write_additions("content", contents)

    result = cli_runner.invoke(
        indexer_cli_group,
        [
            "-C",
            swh_config,
            "journal-client",
            "content_fossology_license",
            "--broker",
            kafka_server,
            "--prefix",
            kafka_prefix,
            "--group-id",
            "test-consumer",
            "--stop-after-objects",
            len(contents),
        ],
        catch_exceptions=False,
    )

    # Check the output
    expected_output = "Done.\n"
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    results = idx_storage.content_fossology_license_get(content_ids)
    assert len(results) == len(expected_results)
    for result in results:
        assert result in expected_results



# Copyright (C) 2016-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
from contextlib import contextmanager
import logging
import os
import shutil
import tempfile
from typing import Any, Dict, Generic, Iterator, List, Optional, Tuple, TypeVar, Union
import warnings

import sentry_sdk
from typing_extensions import TypedDict

from swh.core.config import load_from_envvar, merge_configs
from swh.indexer.storage import INDEXER_CFG_KEY, Sha1, get_indexer_storage
from swh.indexer.storage.interface import IndexerStorageInterface
from swh.model import hashutil
from swh.model.model import Directory, Origin, Sha1Git
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.factory import get_objstorage
from swh.objstorage.interface import objid_from_dict
from swh.storage import get_storage
from swh.storage.interface import StorageInterface


class ObjectsDict(TypedDict, total=False):
    """Typed objects whose keys are names of Kafka topics and values are list of values
    of messages in that topic."""

    content: List[Dict]
    directory: List[Dict]
    origin: List[Dict]
    origin_visit_status: List[Dict]
    raw_extrinsic_metadata: List[Dict]


@contextmanager
def write_to_temp(filename: str, data: bytes, working_directory: str) -> Iterator[str]:
    """Write the sha1's content in a temporary file.

    Args:
        filename: one of sha1's many filenames
        data: the sha1's content to write in temporary
          file
        working_directory: the directory into which the
          file is written

    Returns:
        The path to the temporary file created. That file is
        filled in with the raw content's data.

    """
    os.makedirs(working_directory, exist_ok=True)
    temp_dir = tempfile.mkdtemp(dir=working_directory)
    content_path = os.path.join(temp_dir, filename)

    with open(content_path, "wb") as f:
        f.write(data)

    yield content_path
    shutil.rmtree(temp_dir)


DEFAULT_CONFIG = {
    INDEXER_CFG_KEY: {"cls": "memory"},
    "storage": {"cls": "memory"},
    "objstorage": {"cls": "memory"},
}


TId = TypeVar("TId")
"""type of the ids of index()ed objects."""
TData = TypeVar("TData")
"""type of the objects passed to index()."""
TResult = TypeVar("TResult")
"""return type of index()"""


class BaseIndexer(Generic[TId, TData, TResult], metaclass=abc.ABCMeta):
    """Base class for indexers to inherit from.

    The main entry point is the :func:`run` function which is in
    charge of triggering the computations on the batch dict/ids
    received.

    Indexers can:

    - filter out ids whose data has already been indexed.
    - retrieve ids data from storage or objstorage
    - index this data depending on the object and store the result in
      storage.

    To implement a new object type indexer, inherit from the
    BaseIndexer and implement indexing:

    :meth:`~BaseIndexer.run`:
      object_ids are different depending on object. For example: sha1 for
      content, sha1_git for revision, directory, release, and id for origin

    To implement a new concrete indexer, inherit from the object level
    classes: :class:`ContentIndexer`, :class:`DirectoryIndexer`,
    :class:`OriginIndexer`.

    Then you need to implement the following functions:

    :meth:`~BaseIndexer.filter`:
      filter out data already indexed (in storage).

    :meth:`~BaseIndexer.index_object`:
      compute index on id with data (retrieved from the storage or the
      objstorage by the id key) and return the resulting index computation.

    :meth:`~BaseIndexer.persist_index_computations`:
      persist the results of multiple index computations in the storage.

    The new indexer implementation can also override the following functions:

    :meth:`~BaseIndexer.prepare`:
      Configuration preparation for the indexer.  When overriding, this must
      call the `super().prepare()` instruction.

    :meth:`~BaseIndexer.check`:
      Configuration check for the indexer.  When overriding, this must call the
      `super().check()` instruction.

    :meth:`~BaseIndexer.register_tools`:
      This should return a dict of the tool(s) to use when indexing or
      filtering.

    """

    results: List[TResult]

    USE_TOOLS = True

    catch_exceptions = True
    """Prevents exceptions in `index()` from raising too high. Set to False
    in tests to properly catch all exceptions."""

    storage: StorageInterface
    objstorage: Any
    idx_storage: IndexerStorageInterface

    def __init__(self, config=None, **kw) -> None:
        """Prepare and check that the indexer is ready to run."""
        super().__init__()
        if config is not None:
            self.config = config
        else:
            self.config = load_from_envvar()
        self.config = merge_configs(DEFAULT_CONFIG, self.config)
        self.prepare()
        self.check()
        self.log.debug("%s: config=%s", self, self.config)

    def prepare(self) -> None:
        """Prepare the indexer's needed runtime configuration.
        Without this step, the indexer cannot possibly run.

        """
        config_storage = self.config.get("storage")
        if config_storage:
            self.storage = get_storage(**config_storage)

        self.objstorage = get_objstorage(**self.config["objstorage"])

        idx_storage = self.config[INDEXER_CFG_KEY]
        self.idx_storage = get_indexer_storage(**idx_storage)

        _log = logging.getLogger("requests.packages.urllib3.connectionpool")
        _log.setLevel(logging.WARN)
        self.log = logging.getLogger("swh.indexer")

        if self.USE_TOOLS:
            self.tools = list(self.register_tools(self.config.get("tools", [])))
        self.results = []

    @property
    def tool(self) -> Dict:
        return self.tools[0]

    def check(self) -> None:
        """Check the indexer's configuration is ok before proceeding.
        If ok, does nothing. If not raise error.

        """
        if self.USE_TOOLS and not self.tools:
            raise ValueError("Tools %s is unknown, cannot continue" % self.tools)

    def _prepare_tool(self, tool: Dict[str, Any]) -> Dict[str, Any]:
        """Prepare the tool dict to be compliant with the storage api."""
        return {"tool_%s" % key: value for key, value in tool.items()}

    def register_tools(
        self, tools: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> List[Dict[str, Any]]:
        """Permit to register tools to the storage.

           Add a sensible default which can be overridden if not
           sufficient.  (For now, all indexers use only one tool)

           Expects the self.config['tools'] property to be set with
           one or more tools.

        Args:
            tools: Either a dict or a list of dict.

        Returns:
            list: List of dicts with additional id key.

        Raises:
            ValueError: if not a list nor a dict.

        """
        if isinstance(tools, list):
            tools = list(map(self._prepare_tool, tools))
        elif isinstance(tools, dict):
            tools = [self._prepare_tool(tools)]
        else:
            raise ValueError("Configuration tool(s) must be a dict or list!")

        if tools:
            return self.idx_storage.indexer_configuration_add(tools)
        else:
            return []

    def index(self, id: TId, data: Optional[TData], **kwargs) -> List[TResult]:
        """Index computation for the id and associated raw data.

        Args:
            id: identifier or Dict object
            data: id's data from storage or objstorage depending on
              object type

        Returns:
            dict: a dict that makes sense for the
            :meth:`.persist_index_computations` method.

        """
        raise NotImplementedError()

    def filter(self, ids: List[TId]) -> Iterator[TId]:
        """Filter missing ids for that particular indexer.

        Args:
            ids: list of ids

        Yields:
            iterator of missing ids

        """
        yield from ids

    @abc.abstractmethod
    def persist_index_computations(self, results: List[TResult]) -> Dict[str, int]:
        """Persist the computation resulting from the index.

        Args:

            results: List of results. One result is the
              result of the index function.

        Returns:
            a summary dict of what has been inserted in the storage

        """
        return {}

    def process_journal_objects(self, objects: ObjectsDict) -> Dict:
        """Read swh message objects (content, origin, ...) from the journal to:

        - retrieve the associated objects from the storage backend (e.g. storage,
          objstorage...)
        - execute the associated indexing computations
        - store the results in the indexer storage

        """
        raise NotImplementedError()


class ContentIndexer(BaseIndexer[Sha1, bytes, TResult], Generic[TResult]):
    """A content indexer working on the journal (method `process_journal_objects`) or on
    a list of ids directly (method `run`).

    Note: :class:`ContentIndexer` is not an instantiable object. To
    use it, one should inherit from this class and override the
    methods mentioned in the :class:`BaseIndexer` class.

    """

    def process_journal_objects(self, objects: ObjectsDict) -> Dict:
        """Read content objects from the journal, retrieve their raw content and compute
        content indexing (e.g. mimetype, fossology license, ...).

        Note that once this is deployed, this supersedes the main ContentIndexer.run
        method call and the class ContentPartitionIndexer.
        """
        summary: Dict[str, Any] = {"status": "uneventful"}
        try:
            results = []
            contents = objects.get("content", [])
            content_data = self.objstorage.get_batch(map(objid_from_dict, contents))
            for item, raw_content in zip(contents, content_data):
                id_ = item["sha1"]
                sentry_sdk.set_tag(
                    "swh-indexer-content-sha1", hashutil.hash_to_hex(id_)
                )
                if not raw_content:
                    self.log.warning(
                        "Content %s not found in objstorage",
                        hashutil.hash_to_hex(id_),
                    )
                    continue

                results.extend(self.index(id_, data=raw_content))
        except Exception:
            if not self.catch_exceptions:
                raise
            self.log.exception("Problem when reading contents metadata.")
            sentry_sdk.capture_exception()
            summary["status"] = "failed"
            return summary
        else:
            # Reset tag after we finished processing the given content
            sentry_sdk.set_tag("swh-indexer-content-sha1", "")

        summary_persist = self.persist_index_computations(results)
        self.results = results
        if summary_persist:
            for value in summary_persist.values():
                if value > 0:
                    summary["status"] = "eventful"
            summary.update(summary_persist)
        return summary

    def run(self, ids: List[Sha1], **kwargs) -> Dict:
        """Given a list of ids:

        - retrieve the content from the storage
        - execute the indexing computations
        - store the results

        Args:
            ids (Iterable[Union[bytes, str]]): sha1's identifier list
            **kwargs: passed to the `index` method

        Returns:
            A summary Dict of the task's status

        """
        if "policy_update" in kwargs:
            warnings.warn(
                "'policy_update' argument is deprecated and ignored.",
                DeprecationWarning,
            )
            del kwargs["policy_update"]

        sha1s = [
            hashutil.hash_to_bytes(id_) if isinstance(id_, str) else id_ for id_ in ids
        ]
        results = []
        summary: Dict = {"status": "uneventful"}
        try:
            for sha1 in sha1s:
                sentry_sdk.set_tag(
                    "swh-indexer-content-sha1", hashutil.hash_to_hex(sha1)
                )
                try:
                    raw_content = self.objstorage.get(sha1)
                except ObjNotFoundError:
                    self.log.warning(
                        "Content %s not found in objstorage"
                        % hashutil.hash_to_hex(sha1)
                    )
                    continue
                res = self.index(sha1, raw_content, **kwargs)
                if res:  # If no results, skip it
                    results.extend(res)
                    summary["status"] = "eventful"
            summary = self.persist_index_computations(results)
            self.results = results
        except Exception:
            if not self.catch_exceptions:
                raise
            self.log.exception("Problem when reading contents metadata.")
            sentry_sdk.capture_exception()
            summary["status"] = "failed"
        else:
            # Reset tag after we finished processing the given content
            sentry_sdk.set_tag("swh-indexer-content-sha1", "")
        return summary


class OriginIndexer(BaseIndexer[str, None, TResult], Generic[TResult]):
    """An object type indexer, inherits from the :class:`BaseIndexer` and
    implements Origin indexing using the run method

    Note: the :class:`OriginIndexer` is not an instantiable object.
    To use it in another context one should inherit from this class
    and override the methods mentioned in the :class:`BaseIndexer`
    class.

    """

    def run(self, origin_urls: List[str], **kwargs) -> Dict:
        """Given a list of origin urls:

        - retrieve origins from storage
        - execute the indexing computations
        - store the results

        Args:
            origin_urls: list of origin urls.
            **kwargs: passed to the `index` method

        """
        if "policy_update" in kwargs:
            warnings.warn(
                "'policy_update' argument is deprecated and ignored.",
                DeprecationWarning,
            )
            del kwargs["policy_update"]

        origins = [{"url": url} for url in origin_urls]

        return self.process_journal_objects({"origin": origins})

    def process_journal_objects(self, objects: ObjectsDict) -> Dict:
        """Worker function for ``JournalClient``."""
        origins = [
            Origin(url=status["origin"])
            for status in objects.get("origin_visit_status", [])
            if status["status"] == "full"
        ] + [Origin(url=origin["url"]) for origin in objects.get("origin", [])]

        summary: Dict[str, Any] = {"status": "uneventful"}
        try:
            results = self.index_list(
                origins,
                # no need to check they exist, as we just received either an origin
                # or visit status; which cannot be created by swh-storage unless
                # the origin already exists
                check_origin_known=False,
            )
        except Exception:
            if not self.catch_exceptions:
                raise
            self.log.exception("Problem when processing origins")
            sentry_sdk.capture_exception()
            summary["status"] = "failed"
            return summary

        summary_persist = self.persist_index_computations(results)
        self.results = results
        if summary_persist:
            for value in summary_persist.values():
                if value > 0:
                    summary["status"] = "eventful"
            summary.update(summary_persist)
        return summary

    def index_list(self, origins: List[Origin], **kwargs) -> List[TResult]:
        results = []
        for origin in origins:
            sentry_sdk.set_tag("swh-indexer-origin-url", origin.url)
            results.extend(self.index(origin.url, **kwargs))
        sentry_sdk.set_tag("swh-indexer-origin-url", "")
        return results


class DirectoryIndexer(BaseIndexer[Sha1Git, Directory, TResult], Generic[TResult]):
    """An object type indexer, inherits from the :class:`BaseIndexer` and
    implements Directory indexing using the run method

    Note: the :class:`DirectoryIndexer` is not an instantiable object.
    To use it in another context one should inherit from this class
    and override the methods mentioned in the :class:`BaseIndexer`
    class.

    """

    def run(self, ids: List[Sha1Git], **kwargs) -> Dict:
        """Given a list of sha1_gits:

        - retrieve directories from storage
        - execute the indexing computations
        - store the results

        Args:
            ids: sha1_git's identifier list

        """
        if "policy_update" in kwargs:
            warnings.warn(
                "'policy_update' argument is deprecated and ignored.",
                DeprecationWarning,
            )
            del kwargs["policy_update"]

        directory_ids = [
            hashutil.hash_to_bytes(id_) if isinstance(id_, str) else id_ for id_ in ids
        ]

        return self._process_directories([(dir_id, None) for dir_id in directory_ids])

    def process_journal_objects(self, objects: ObjectsDict) -> Dict:
        """Worker function for ``JournalClient``."""
        return self._process_directories(
            [
                (dir_["id"], Directory.from_dict(dir_))
                for dir_ in objects.get("directory", [])
            ]
        )

    def _process_directories(
        self,
        directories: Union[List[Tuple[Sha1Git, Directory]], List[Tuple[Sha1Git, None]]],
    ) -> Dict:
        summary: Dict[str, Any] = {"status": "uneventful"}
        results = []

        # TODO: fetch raw_manifest when useful?

        for dir_id, dir_ in directories:
            swhid = f"swh:1:dir:{hashutil.hash_to_hex(dir_id)}"
            sentry_sdk.set_tag("swh-indexer-directory-swhid", swhid)
            try:
                results.extend(self.index(dir_id, dir_))
            except Exception:
                if not self.catch_exceptions:
                    raise
                self.log.exception("Problem when processing directory")
                sentry_sdk.capture_exception()
                summary["status"] = "failed"
            else:
                sentry_sdk.set_tag("swh-indexer-directory-swhid", "")

        summary_persist = self.persist_index_computations(results)
        if summary_persist:
            for value in summary_persist.values():
                if value > 0:
                    summary["status"] = "eventful"
            summary.update(summary_persist)
        self.results = results
        return summary



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import json
import sys
from typing import Any, Dict, List, Optional
import uuid

from pybtex.database import Entry, Person
import rdflib

from swh.indexer.codemeta import compact, expand
from swh.indexer.metadata_dictionary.cff import CffMapping
from swh.indexer.namespaces import RDF, SCHEMA, SPDX_LICENSES
from swh.model.swhids import ObjectType, QualifiedSWHID

TMP_ROOT_URI_PREFIX = "https://www.softwareheritage.org/schema/2022/indexer/tmp-node/"
"""IRI used for `skolemization <https://www.w3.org/TR/rdf11-concepts/#section-skolemization>`_;
it is not used outside :func:`codemeta_to_bibtex`.
"""


def codemeta_to_bibtex(
    doc: Dict[str, Any], swhid: Optional[QualifiedSWHID] = None
) -> str:
    doc = compact(doc, False)

    identifiers = []

    if "id" in doc:
        identifiers.append(doc["id"])
    else:
        doc["id"] = f"_:{uuid.uuid4()}"

    id_: rdflib.term.Node
    if doc["id"].startswith("_:"):
        id_ = rdflib.term.BNode(doc["id"][2:])
    else:
        # using a base in case the id is not an absolute URI
        id_ = rdflib.term.URIRef(doc["id"], base=TMP_ROOT_URI_PREFIX)
        doc["id"] = str(id_)

    # workaround for https://github.com/codemeta/codemeta/pull/322
    if "identifier" in doc:
        if isinstance(doc["identifier"], list):
            for identifier in doc["identifier"]:
                if isinstance(identifier, str) and "/" not in identifier:
                    identifiers.append(identifier)
        elif isinstance(doc["identifier"], str) and "/" not in doc["identifier"]:
            identifiers.append(doc["identifier"])

    doc = expand(doc)
    g = rdflib.Graph().parse(data=json.dumps(doc), format="json-ld")
    persons: Dict[str, List[Person]] = collections.defaultdict(list)
    fields: Dict[str, Any] = {}

    def add_person(persons: List[Person], person_id: rdflib.term.Node) -> None:
        person = Person()
        for _, _, name in g.triples((person_id, SCHEMA.name, None)):
            if (person_id, RDF.type, SCHEMA.Organization) in g:
                # prevent interpreting the name as "Firstname Lastname" and reformatting
                # it to "Lastname, Firstname"
                person.last_names.append(name)
            else:
                person = Person(name)

        for _, _, given_name in g.triples((person_id, SCHEMA.givenName, None)):
            person.first_names.append(given_name)

        for _, _, family_name in g.triples((person_id, SCHEMA.familyName, None)):
            person.last_names.append(family_name)

        if str(person) and person not in persons:
            persons.append(person)

    def add_affiliations(person: rdflib.term.Node) -> None:
        for _, _, organization in g.triples((person, SCHEMA.affiliation, None)):
            add_person(persons["organization"], organization)

    # abstract
    for _, _, description in g.triples((id_, SCHEMA.description, None)):
        fields["abstract"] = description
        break

    # authors, which are an ordered list
    for _, _, author_list in g.triples((id_, SCHEMA.author, None)):
        if author_list == RDF.nil:
            # Workaround for https://github.com/RDFLib/rdflib/pull/2818
            continue
        for author in rdflib.collection.Collection(g, author_list):
            add_person(persons["author"], author)
            add_affiliations(author)

    # date
    for _, _, date in g.triples((id_, SCHEMA.datePublished, None)):
        fields["date"] = date
        break
    else:
        for _, _, date in g.triples((id_, SCHEMA.dateCreated, None)):
            fields["date"] = date
            break
        else:
            for _, _, date in g.triples((id_, SCHEMA.dateModified, None)):
                fields["date"] = date
                break
    if "date" in fields:
        (fields["year"], fields["month"], _) = fields["date"].split("-")

    # identifier, doi, hal_id
    entry_key = None
    for _, _, identifier in g.triples((id_, SCHEMA.identifier, None)):
        identifiers.append(identifier)
    for identifier in identifiers:
        if entry_key is None and "/" not in identifier:
            # Avoid URLs
            entry_key = identifier
        if identifier.startswith("https://doi.org/"):
            fields["doi"] = identifier
        if identifier.startswith("hal-"):
            fields["hal_id"] = identifier

    # editor
    for _, _, editor in g.triples((id_, SCHEMA.editor, None)):
        add_person(persons["editor"], editor)
        add_affiliations(editor)

    # file
    for _, _, download_url in g.triples((id_, SCHEMA.downloadUrl, None)):
        fields["file"] = download_url
        break

    # license (represented by "Person" as it's the only way to make pybtex format
    # them as a list)
    for _, _, license in g.triples((id_, SCHEMA.license, None)):
        if license is None:
            continue
        license_ = str(license)
        if license_.startswith(str(SPDX_LICENSES)):
            license_ = license_[len(str(SPDX_LICENSES)) :]
            if license_.endswith(".html"):
                license_ = license_[:-5]
            persons["license"].append(Person(last=license_))

    # publisher
    for _, _, publisher in g.triples((id_, SCHEMA.publisher, None)):
        add_person(persons["publisher"], publisher)
        add_affiliations(publisher)

    # repository
    for _, _, code_repository in g.triples((id_, SCHEMA.codeRepository, None)):
        fields["repository"] = code_repository
        break

    # title
    for _, _, name in g.triples((id_, SCHEMA.name, None)):
        fields["title"] = name
        break

    # url
    for _, _, name in g.triples((id_, SCHEMA.url, None)):
        fields["url"] = name
        break

    # version
    for _, _, version in g.triples((id_, SCHEMA.softwareVersion, None)):
        fields["version"] = version
        break
    else:
        for _, _, version in g.triples((id_, SCHEMA.version, None)):
            fields["version"] = version

    # entry_type
    if swhid:
        fields["swhid"] = str(swhid)
        if swhid.object_type == ObjectType.SNAPSHOT:
            entry_type = "software"
        elif swhid.object_type == ObjectType.CONTENT:
            entry_type = "codefragment"
        else:
            entry_type = "softwareversion"
    elif "version" in fields:
        entry_type = "softwareversion"
    else:
        entry_type = "software"

    entry = Entry(
        entry_type,
        persons=persons,
        fields=fields,
    )

    entry.key = entry_key or "REPLACEME"
    return entry.to_string(bib_format="bibtex")


def cff_to_bibtex(content: str, swhid: Optional[QualifiedSWHID] = None) -> str:
    codemeta = CffMapping().translate(raw_content=content.encode("utf-8"))
    if codemeta is None:
        codemeta = {}
    return codemeta_to_bibtex(codemeta, swhid)


if __name__ == "__main__":
    for filename in sys.argv[1:]:
        if filename == "-":
            print(codemeta_to_bibtex(json.load(sys.stdin)))
        else:
            with open(filename) as f:
                print(codemeta_to_bibtex(json.load(f)))
